package models

import (
	"errors"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"github.com/shopspring/decimal"
	"sort"
	"strconv"
	"strings"
	"time"
)

type EdbInfoData struct {
	EdbDataId int       `description:"数据ID"`
	DataTime  time.Time `description:"数据日期"`
	Value     float64   `description:"数据"`
}

type BaseCalculateBySearchData struct {
	DataList      []*EdbInfoSearchData
	Frequency     string `description:"需要转换的频度"`
	Formula       interface{}
	Calendar      string `description:"公历/农历"`
	MoveType      int    `description:"移动方式:1:领先(默认),2:滞后"`
	MoveFrequency string `description:"移动频度"`
	FromFrequency string `description:"来源的频度"`
	Source        int    `description:"1：累计值转月；2：累计值转季；3：同比值；4：同差值；5：N数值移动平均数计算；6：环比值；7：环差值；8：升频；9：降频；10：时间移位；11：超季节性；12：年化；13：累计值；14：累计值年初至今；15：指数修匀；16：日均值"`
}

type BaseCalculate struct {
	DataList      []*EdbInfoData
	Frequency     string `description:"需要转换的频度"`
	Formula       interface{}
	Calendar      string `description:"公历/农历"`
	MoveType      int    `description:"移动方式:1:领先(默认),2:滞后"`
	MoveFrequency string `description:"移动频度"`
	FromFrequency string `description:"来源的频度"`
	Source        int    `description:"1：累计值转月；2：累计值转季；3：同比值；4：同差值；5：N数值移动平均数计算；6：环比值；7：环差值；8：升频；9：降频；10：时间移位；11：超季节性；12：年化；13：累计值；14：累计值年初至今；15：指数修匀；16：日均值"`
}

type BaseCalculateResp struct {
	DataMap  map[string]float64
	DateList []string
}

func (obj BaseCalculate) Ljzzy() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)

	dataList := obj.DataList // 升序
	yearMap := make(map[int]map[int]*EdbInfoData)
	dataLen := len(dataList)
	for i := 0; i < dataLen; i++ {
		item := dataList[i]
		itemDate := item.DataTime
		year := itemDate.Year()
		month := int(itemDate.Month())
		if monthMap, yok := yearMap[year]; yok {
			monthMap[month] = item
			yearMap[year] = monthMap
		} else {
			monthMap = make(map[int]*EdbInfoData)
			monthMap[month] = item
			yearMap[year] = monthMap
		}
	}

	for yk, yv := range yearMap {
		_, oneMonthOk := yv[1]
		_, twoMonthOk := yv[2]
		if !oneMonthOk && !twoMonthOk {
			continue
		}
		for i := 1; i <= 12; i++ {
			dataCurrentItem := yv[i]
			if dataCurrentItem == nil {
				continue
			}
			var date time.Time
			var val float64

			if i == 1 || i == 2 {
				if _, mok := yv[1]; mok { //1月有值
					if i == 1 {
						date = dataCurrentItem.DataTime
						val, _ = decimal.NewFromFloat(dataCurrentItem.Value).Float64() //a.Div(b).Float64()
					}
					if i == 2 {
						dataOneItem := yv[1]
						date = dataCurrentItem.DataTime
						twoMonth := decimal.NewFromFloat(dataCurrentItem.Value)
						oneMonth := decimal.NewFromFloat(dataOneItem.Value)
						val, _ = twoMonth.Sub(oneMonth).Float64()
					}
				} else { //1月无值
					dataTwoItem := yv[2]
					if i == 1 {
						dateStr := strconv.Itoa(yk) + "-01-31"
						a := decimal.NewFromFloat(dataTwoItem.Value)
						b := decimal.NewFromFloat(2.0)
						val, _ = a.Div(b).Float64()

						date, err = time.ParseInLocation(utils.FormatDate, dateStr, time.Local)
						if err != nil {
							return
						}

					}
					if i == 2 {
						{
							dateStr := strconv.Itoa(yk) + "-01-31"
							a := decimal.NewFromFloat(dataTwoItem.Value)
							b := decimal.NewFromFloat(2.0)
							val, _ = a.Div(b).Float64()
							date, err = time.ParseInLocation(utils.FormatDate, dateStr, time.Local)
							if err != nil {
								return
							}
							dateDataMap[date] = val
						}

						date = dataCurrentItem.DataTime
						a := decimal.NewFromFloat(dataTwoItem.Value)
						b := decimal.NewFromFloat(2.0)
						val, _ = a.Div(b).Float64()
					}
				}
			} else {
				dataPreItem := yv[i-1]
				if dataCurrentItem != nil && dataPreItem != nil {
					date = dataCurrentItem.DataTime
					a := decimal.NewFromFloat(dataCurrentItem.Value)
					b := decimal.NewFromFloat(dataPreItem.Value)
					val, _ = a.Sub(b).Float64()
				}
			}

			if date.IsZero() {
				continue
			}

			dateDataMap[date] = val
		}
	}

	return
}

func (obj BaseCalculate) Ljzzj() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList // 升序
	yearMap := make(map[int]map[int]*EdbInfoData)
	dataLen := len(dataList)
	for i := 0; i < dataLen; i++ {
		item := dataList[i]
		itemDate := item.DataTime
		year := itemDate.Year()
		quarter := utils.MonthQuarterMap[int(itemDate.Month())]
		if quarterMap, yok := yearMap[year]; yok {
			quarterMap[quarter] = item
			yearMap[year] = quarterMap
		} else {
			quarterMap = make(map[int]*EdbInfoData)
			quarterMap[quarter] = item
			yearMap[year] = quarterMap
		}
	}

	for yk, yv := range yearMap {
		_, oneQuarterOk := yv[1]
		_, twoQuarterOk := yv[2]
		if !oneQuarterOk && !twoQuarterOk {
			continue
		}
		for i := 1; i <= 4; i++ {
			dataCurrentItem := yv[i]
			if dataCurrentItem == nil {
				continue
			}
			var date time.Time
			var val float64

			if i == 1 || i == 2 {
				if _, mok := yv[1]; mok { //第1个季度有值
					if i == 1 {
						date = dataCurrentItem.DataTime
						val, _ = decimal.NewFromFloat(dataCurrentItem.Value).Float64() //a.Div(b).Float64()
					}
					if i == 2 {
						dataOneItem := yv[1]
						date = dataCurrentItem.DataTime
						twoQuarter := decimal.NewFromFloat(dataCurrentItem.Value)
						oneQuarter := decimal.NewFromFloat(dataOneItem.Value)
						val, _ = twoQuarter.Sub(oneQuarter).Float64()
					}
				} else { //第1个季度无值
					dataTwoItem := yv[2]
					if i == 1 {
						dateStr := strconv.Itoa(yk) + "-03-31"
						a := decimal.NewFromFloat(dataTwoItem.Value)
						b := decimal.NewFromFloat(2.0)
						val, _ = a.Div(b).Float64()
						date, err = time.ParseInLocation(utils.FormatDate, dateStr, time.Local)
						if err != nil {
							return
						}
					}
					if i == 2 {
						{
							dateStr := strconv.Itoa(yk) + "-03-31"
							a := decimal.NewFromFloat(dataTwoItem.Value)
							b := decimal.NewFromFloat(2.0)
							val, _ = a.Div(b).Float64()
							date, err = time.ParseInLocation(utils.FormatDate, dateStr, time.Local)
							if err != nil {
								return
							}

							dateDataMap[date] = val
						}

						date = dataCurrentItem.DataTime
						a := decimal.NewFromFloat(dataTwoItem.Value)
						b := decimal.NewFromFloat(2.0)
						val, _ = a.Div(b).Float64()
					}
				}
			} else {
				dataPreItem := yv[i-1]
				if dataCurrentItem != nil && dataPreItem != nil {
					date = dataCurrentItem.DataTime
					a := decimal.NewFromFloat(dataCurrentItem.Value)
					b := decimal.NewFromFloat(dataPreItem.Value)
					val, _ = a.Sub(b).Float64()
				}
			}

			if date.IsZero() {
				continue
			}

			dateDataMap[date] = val
		}
	}

	return
}

func (obj BaseCalculate) Tbz() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	frequency := obj.FromFrequency

	dataList = reverseSliceByDesc(dataList)

	var dateArr []time.Time
	dataMap := make(map[string]*EdbInfoData) // 避免因为时间戳导致的日期不对，还是用string来表示比较合适
	var firstDataTime time.Time              // 第一条数据的日期
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime.Format(utils.FormatDate)] = v
		firstDataTime = v.DataTime
	}

	if firstDataTime.IsZero() {
		errMsg = "无数据"
		err = errors.New(errMsg)
		return
	}
	firstDataTime = firstDataTime.AddDate(1, 0, 0)

	for _, currentDate := range dateArr {
		if currentDate.Before(firstDataTime) {
			break
		}

		currentDateStr := currentDate.Format(utils.FormatDate)
		currentItem, ok := dataMap[currentDateStr]
		if !ok {
			continue
		}

		var findItem *EdbInfoData

		preDate := currentDate.AddDate(-1, 0, 0)
		preDateStr := preDate.Format(utils.FormatDate)
		if findItem, ok = dataMap[preDateStr]; !ok { //上一年同期没找到
			if frequency == "月度" { //向上和向下，各找一个月
				for i := 0; i <= 35; i++ {
					nextDateDay := preDate.AddDate(0, 0, i)
					nextDateDayStr := nextDateDay.Format(utils.FormatDate)
					if findItem, ok = dataMap[nextDateDayStr]; ok { //上一年同期->下一个月找到
						break
					} else {
						preDateDay := preDate.AddDate(0, 0, -i)
						preDateDayStr := preDateDay.Format(utils.FormatDate)
						if findItem, ok = dataMap[preDateDayStr]; ok { //上一年同期->上一个月找到
							break
						}
					}
				}
			} else if frequency == "季度" || frequency == "年度" {
				if findItem, ok = dataMap[preDateStr]; ok { //上一年同期->下一个月找到
					break
				}
			} else {
				nextDateDay := preDate.AddDate(0, 0, 1)
				nextDateDayStr := nextDateDay.Format(utils.FormatDate)

				preDateDay := preDate.AddDate(0, 0, -1)
				preDateDayStr := preDateDay.Format(utils.FormatDate)

				for i := 0; i < 35; i++ {
					if i >= 1 {
						nextDateDay = nextDateDay.AddDate(0, 0, i)
						nextDateDayStr = nextDateDay.Format(utils.FormatDate)
					}
					if findItem, ok = dataMap[nextDateDayStr]; ok { //上一年同期->下一个月找到
						break
					} else {
						if i >= 1 {
							preDateDay = preDateDay.AddDate(0, 0, -i)
							preDateDayStr = preDateDay.Format(utils.FormatDate)
						}
						if findItem, ok = dataMap[preDateDayStr]; ok { //上一年同期->上一个月找到
							break
						}
					}
				}
			}

		}

		if findItem == nil {
			continue
		}

		dateDataMap[currentDate] = tbzDiv(currentItem.Value, findItem.Value)
	}

	return
}

func tbzDiv(a, b float64) float64 {
	var val float64
	if b != 0 {
		af := decimal.NewFromFloat(a)
		bf := decimal.NewFromFloat(b)
		val, _ = af.Div(bf).Sub(decimal.NewFromFloat(1)).Round(4).Float64()
	}
	return val
}

func (obj BaseCalculate) Tcz() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList // 降序
	frequency := obj.FromFrequency

	dataList = reverseSliceByDesc(dataList)
	var dateArr []time.Time
	dataMap := make(map[string]*EdbInfoData)
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime.Format(utils.FormatDate)] = v
	}

	for _, currentDate := range dateArr {
		currentItem, ok := dataMap[currentDate.Format(utils.FormatDate)]
		if !ok {
			continue
		}

		var findItem *EdbInfoData

		preDate := currentDate.AddDate(-1, 0, 0)
		preDateStr := preDate.Format(utils.FormatDate)
		if findItem, ok = dataMap[preDateStr]; !ok {
			if frequency == "月度" { //向上和向下，各找一个月
				for i := 0; i <= 35; i++ {
					nextDateDay := preDate.AddDate(0, 0, i)
					nextDateDayStr := nextDateDay.Format(utils.FormatDate)
					if findItem, ok = dataMap[nextDateDayStr]; ok { //上一年同期->下一个月找到
						break
					} else {
						preDateDay := preDate.AddDate(0, 0, -i)
						preDateDayStr := preDateDay.Format(utils.FormatDate)
						if findItem, ok = dataMap[preDateDayStr]; ok { //上一年同期->上一个月找到
							break
						}
					}
				}
			} else if frequency == "季度" || frequency == "年度" {
				if findItem, ok = dataMap[preDateStr]; ok { //上一年同期->下一个月找到
					break
				}
			} else {
				for i := 0; i < 35; i++ {
					nextDateDay := preDate.AddDate(0, 0, i)
					nextDateDayStr := nextDateDay.Format(utils.FormatDate)
					if findItem, ok = dataMap[nextDateDayStr]; ok { //上一年同期->下一个月找到
						break
					} else {
						preDateDay := preDate.AddDate(0, 0, -i)
						preDateDayStr := preDateDay.Format(utils.FormatDate)
						if findItem, ok = dataMap[preDateDayStr]; ok { //上一年同期->上一个月找到
							break
						}
					}
				}
			}
		}

		if findItem == nil {
			continue
		}

		dateDataMap[currentDate] = tczSub(currentItem.Value, findItem.Value)
	}

	return
}

func tczSub(a, b float64) float64 {
	af := decimal.NewFromFloat(a)
	bf := decimal.NewFromFloat(b)
	val, _ := af.Sub(bf).Round(4).Float64()
	return val
}

func (obj BaseCalculate) Nszydpjjs() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	fromDataList := obj.DataList
	nVal := fmt.Sprint(obj.Formula)
	formulaInt, err := strconv.Atoi(nVal)
	if err != nil {
		return
	}


	fromDataList = reverseSliceByDesc(fromDataList)

	var fromDateArr []time.Time
	fromDataMap := make(map[time.Time]*EdbInfoData)
	for _, v := range fromDataList {
		fromDateArr = append(fromDateArr, v.DataTime)
		fromDataMap[v.DataTime] = v
	}

	arrLen := len(fromDateArr)
	for ak, currentDate := range fromDateArr {
		var valArr []float64
		if findItem, ok := fromDataMap[currentDate]; ok {
			valArr = append(valArr, findItem.Value)
		} else {
			continue
		}
		if ak+1 != arrLen {
			for i := 1; i < formulaInt; i++ {
				arrIndex := ak + i
				if arrIndex >= arrLen {
					break
				}
				arrVal := fromDateArr[arrIndex]
				if findItem, ok := fromDataMap[arrVal]; ok {
					valArr = append(valArr, findItem.Value)
				} else {
					continue
				}
			}
		}
		valArrLen := len(valArr)
		totalVal := decimal.NewFromFloat(0.00)
		for _, v := range valArr {
			newDecimal := decimal.NewFromFloat(v)
			totalVal = totalVal.Add(newDecimal)
		}
		af := totalVal //decimal.NewFromFloat(totalVal)
		bf := decimal.NewFromFloat(float64(valArrLen))
		val, _ := af.Div(bf).Round(4).Float64()

		dateDataMap[currentDate] = val
	}

	return
}

func (obj BaseCalculate) Hbz() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	nVal := fmt.Sprint(obj.Formula)
	formulaInt, err := strconv.Atoi(nVal)
	if err != nil {
		return
	}

	dataList = reverseSliceByDesc(dataList)

	var dateArr []time.Time
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
	}

	dataLen := len(dataList)
	for i := 0; i < dataLen; i++ {
		j := i + formulaInt
		if j < dataLen {
			currentItem := dataList[i]
			preItem := dataList[j]

			if currentItem == nil || preItem == nil {
				continue
			}

			val, ok := hbzDiv(currentItem.Value, preItem.Value)
			if !ok {
				continue
			}

			dateDataMap[currentItem.DataTime] = val
		}
	}

	return
}

func hbzDiv(current, pre float64) (val float64, ok bool) {
	if pre == 0 {
		return
	}
	currentVal := decimal.NewFromFloat(current)
	preVal := decimal.NewFromFloat(pre)
	val, _ = currentVal.Sub(preVal).Div(preVal).Round(4).Float64()
	ok = true

	return
}

func (obj BaseCalculate) Hcz() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	nVal := fmt.Sprint(obj.Formula)
	formulaInt, err := strconv.Atoi(nVal)
	if err != nil {
		return
	}

	dataList = reverseSliceByDesc(dataList)

	var dateArr []time.Time
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
	}

	dataLen := len(dataList)
	for i := 0; i < dataLen; i++ {
		j := i + formulaInt
		if j < dataLen {
			currentItem := dataList[i]
			preItem := dataList[j]

			if currentItem == nil || preItem == nil {
				continue
			}

			val := hczDiv(currentItem.Value, preItem.Value)

			dateDataMap[currentItem.DataTime] = val
		}
	}

	return
}

func hczDiv(current, pre float64) float64 {
	currentVal := decimal.NewFromFloat(current)
	preVal := decimal.NewFromFloat(pre)
	val, _ := currentVal.Sub(preVal).Round(4).Float64()

	return val
}

func (obj BaseCalculate) UpFrequency() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList

	var dateArr []time.Time
	dataMap := make(map[time.Time]*EdbInfoData)
	fromDataMap := make(map[time.Time]float64)
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
		fromDataMap[v.DataTime] = v.Value
	}

	if obj.FromFrequency == `日度` {
		errMsg = `不能选择日度指标`
		err = errors.New(errMsg)
		return
	}

	nowDay := time.Now()
	nowDay = time.Date(nowDay.Year(), nowDay.Month(), nowDay.Day(), 0, 0, 0, 0, nowDay.Location())

	dataLen := len(dataList)
	lastI := dataLen - 1
	for i := 0; i < dataLen; i++ {
		currentItem := dataList[i]
		currentDate := currentItem.DataTime
		var day int
		var preItem *EdbInfoData
		var preDate time.Time

		if i == 0 {
			dateDataMap[currentItem.DataTime] = currentItem.Value
			continue
		}

		j := i - 1

		preItem = dataList[j]
		preDate = preItem.DataTime
		day = int(currentDate.Sub(preItem.DataTime).Hours() / float64(24))

		for k := 1; k < day; k++ {
			needDay := preDate.AddDate(0, 0, k)
			dateDataMap[needDay] = preItem.Value
		}
		dateDataMap[currentItem.DataTime] = currentItem.Value

		if i == lastI {
			day = int(nowDay.Sub(currentDate).Hours() / float64(24))

			for k := 1; k <= day; k++ {
				needDay := currentDate.AddDate(0, 0, k)
				dateDataMap[needDay] = currentItem.Value
			}
		}

	}

	return
}

func (obj BaseCalculate) DownFrequency() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	edbFrequency := obj.Frequency
	formula := obj.Formula.(string)

	if !CheckFrequency(obj.FromFrequency, obj.Frequency) {
		errMsg = "频度异常,不允许低频降频到高频"
		err = errors.New(errMsg)
		return
	}

	var dateArr []time.Time
	dataMap := make(map[time.Time]*EdbInfoData)
	fromDataMap := make(map[time.Time]float64)
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
		fromDataMap[v.DataTime] = v.Value
	}

	dataLen := len(dataList)
	if dataLen <= 0 {
		return
	}
	startDataTime := dataList[0].DataTime
	endDataTime := dataList[dataLen-1].DataTime

	var lastValue float64     // 最近的值
	var nextEndDate time.Time // 下一个节点的日期
	weekDayDataList := make([]float64, 0)
	for tmpStartDataTime := startDataTime; !tmpStartDataTime.After(endDataTime); tmpStartDataTime = tmpStartDataTime.AddDate(0, 0, 1) {
		if tmpData, ok := dataMap[tmpStartDataTime]; ok {
			tmpValue := decimal.NewFromFloat(tmpData.Value)
			tmpValueFloat, _ := tmpValue.Round(4).Float64()
			weekDayDataList = append(weekDayDataList, tmpValueFloat)
		}
		if nextEndDate.IsZero() {
			nextEndDate = utils.GetFrequencyEndDay(tmpStartDataTime, edbFrequency)
		}

		switch edbFrequency {
		case "周度":
			if tmpStartDataTime.Weekday() != 5 {
				continue
			} else {
				nextEndDate = tmpStartDataTime.AddDate(0, 0, 7)
			}
		case "旬度":
			nextDay := tmpStartDataTime.AddDate(0, 0, 1)
			if nextDay.Day() != 1 && nextDay.Day() != 11 && nextDay.Day() != 21 {
				continue
			} else {
				if nextDay.Day() == 1 || nextDay.Day() == 11 {
					nextEndDate = nextDay.AddDate(0, 0, 9)
				} else {
					tmpNextMonth := nextDay.AddDate(0, 1, 0)
					nextEndDate = time.Date(tmpNextMonth.Year(), tmpNextMonth.Month(), 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
				}
			}
		case "月度":
			nextDay := tmpStartDataTime.AddDate(0, 0, 1)
			if nextDay.Day() != 1 {
				continue
			} else {
				nextEndDate = nextDay.AddDate(0, 1, -1)
			}
		case "季度":
			nextDay := tmpStartDataTime.AddDate(0, 0, 1)
			if (nextDay.Month() == 1 || nextDay.Month() == 4 || nextDay.Month() == 7 || nextDay.Month() == 10) && nextDay.Day() == 1 {
				nextEndDate = nextDay.AddDate(0, 3, -1)
			} else {
				continue
			}
		case "年度":
			if tmpStartDataTime.Month() == 12 && tmpStartDataTime.Day() == 31 {
				nextEndDate = tmpStartDataTime.AddDate(1, 0, 0)
			} else {
				continue
			}
		default:
			err = errors.New("错误的频度:" + edbFrequency)
			return
		}

		var currVal float64
		lenWeekDayDataList := len(weekDayDataList)
		if len(weekDayDataList) <= 0 {
			currVal = lastValue
		} else {
			if formula == "期末值" {
				currVal = weekDayDataList[lenWeekDayDataList-1]
			} else {
				sumValDeci := decimal.NewFromFloat(0)
				for _, v := range weekDayDataList {
					tmpValDeci := decimal.NewFromFloat(v)
					sumValDeci = sumValDeci.Add(tmpValDeci)
				}
				lenDeci := decimal.NewFromInt(int64(lenWeekDayDataList))
				currVal, _ = sumValDeci.Div(lenDeci).Round(4).Float64()
			}
		}

		dateDataMap[tmpStartDataTime] = currVal

		weekDayDataList = make([]float64, 0)
	}

	if len(weekDayDataList) > 0 {
		var currVal float64
		lenWeekDayDataList := len(weekDayDataList)
		if len(weekDayDataList) < 0 {
			currVal = lastValue
		} else {
			if formula == "期末值" {
				currVal = weekDayDataList[lenWeekDayDataList-1]
			} else {
				sumValDeci := decimal.NewFromFloat(0)
				for _, v := range weekDayDataList {
					tmpValDeci := decimal.NewFromFloat(v)
					sumValDeci = sumValDeci.Add(tmpValDeci)
				}
				lenDeci := decimal.NewFromInt(int64(lenWeekDayDataList))
				currVal, _ = sumValDeci.Div(lenDeci).Round(4).Float64()
			}
		}

		dateDataMap[nextEndDate] = currVal
	}

	return
}

func (obj BaseCalculate) TimeShift() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList

	nVal := fmt.Sprint(obj.Formula)
	formulaInt, err := strconv.Atoi(nVal)
	if err != nil {
		return
	}

	moveType := obj.MoveType
	moveFrequency := obj.MoveFrequency

	var dateArr []time.Time
	dataMap := make(map[time.Time]*EdbInfoData)
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
	}

	var shiftDay int
	switch moveFrequency {
	case "天":
		shiftDay = formulaInt
	case "周":
		shiftDay = formulaInt * 7
	case "月":
		shiftDay = formulaInt * 30
	case "季":
		shiftDay = formulaInt * 90
	case "年":
		shiftDay = formulaInt * 365
	default:
		shiftDay = formulaInt
	}

	if moveType == 2 {
		shiftDay = -shiftDay
	}

	dataLen := len(dataList)
	for i := 0; i < dataLen; i++ {
		currentItem := dataList[i]
		currentDate := currentItem.DataTime

		newDate := currentDate.AddDate(0, 0, shiftDay)

		val, _ := decimal.NewFromFloat(currentItem.Value).Round(4).Float64()

		dateDataMap[newDate] = val
	}

	return
}

func (obj BaseCalculate) Cjjx() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList

	nVal := fmt.Sprint(obj.Formula)
	formulaInt, err := strconv.Atoi(nVal)
	if err != nil {
		return
	}

	calendar := obj.Calendar

	var dateArr []time.Time
	dataMap := make(map[time.Time]*EdbInfoData)
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
	}

	handleDataMap := make(map[time.Time]float64)
	_, err = HandleDataByLinearRegressionByTime(dataList, handleDataMap)
	if err != nil {
		return
	}

	moveDayMap := make(map[int]int, 0) // 每个年份的春节公历
	var lastDataDay time.Time
	if len(dataList) > 0 {
		lastDataDay = dataList[0].DataTime
	}

	for _, currentDate := range dateArr {
		if strings.Contains(currentDate.Format(utils.FormatDate), "02-29") {
			continue
		}

		if calendar == "农历" && currentDate.Month() > 5 && currentDate.Month() < 11 {
			continue
		}

		currentItem, ok := dataMap[currentDate]
		if !ok {
			continue
		}

		pastValueList := make([]float64, 0)                      // 过去几期的数据
		pastValueList = append(pastValueList, currentItem.Value) //当前日期

		for i := 1; i < formulaInt; i++ {
			historyPreDate := currentDate.AddDate(-i, 0, 0)
			moveDay := 0
			if calendar == "农历" {
				if tmpMoveDay, ok := moveDayMap[historyPreDate.Year()]; !ok {
					moveDay, err = getMoveDay(lastDataDay, historyPreDate)
					if err != nil {
						return
					}
				} else {
					moveDay = tmpMoveDay
				}

				historyPreDate = historyPreDate.AddDate(0, 0, moveDay)
			}

			if tmpValue, ok := handleDataMap[historyPreDate]; ok { //上一年同期找到
				pastValueList = append(pastValueList, tmpValue)
			}
		}

		if len(pastValueList) == formulaInt {
			val, ok := cjjxSub(currentItem.Value, pastValueList)
			if !ok {
				continue
			}
			dateDataMap[currentDate] = val
		}
	}

	return
}

func cjjxSub(currValue float64, pastValue []float64) (value float64, ok bool) {
	num := len(pastValue)
	if num == 0 {
		return
	}
	numDecimal := decimal.NewFromInt(int64(num))

	af := decimal.NewFromFloat(currValue)
	bf := decimal.NewFromFloat(pastValue[0])

	for k := 1; k < num; k++ {
		tmpVal := decimal.NewFromFloat(pastValue[k])
		bf = bf.Add(tmpVal)
	}
	value, _ = af.Sub(bf.Div(numDecimal)).Round(4).Float64()
	ok = true

	return
}

func (obj BaseCalculate) Annualized() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	fromDataList := obj.DataList
	lenFromDataList := len(fromDataList)
	if lenFromDataList <= 0 {
		return
	}

	handleDataMap := make(map[time.Time]float64)
	_, err = HandleDataByLinearRegressionByTime(fromDataList, handleDataMap)
	if err != nil {
		return
	}

	yearLastValMap := make(map[int]float64)
	startDataTime := fromDataList[0].DataTime
	endDataTime := fromDataList[lenFromDataList-1].DataTime
	for i := startDataTime.Year(); i <= endDataTime.Year(); i++ {
		tmpDate := time.Date(i, 12, 31, startDataTime.Hour(), startDataTime.Minute(), startDataTime.Second(), startDataTime.Nanosecond(), time.Local)
		if tmpVal, ok := handleDataMap[tmpDate]; ok {
			yearLastValMap[i] = tmpVal
		}
	}

	for _, v := range fromDataList {
		currDate := v.DataTime

		perValMap := make(map[time.Time]float64)
		for i := 1; i <= 3; i++ {
			tmpDateTime := currDate.AddDate(-i, 0, 0)
			if tmpVal, ok := handleDataMap[tmpDateTime]; ok {
				perValMap[tmpDateTime] = tmpVal
			}
		}
		lenPerValMap := len(perValMap)
		if lenPerValMap < 2 {
			continue
		}

		divValList := make([]decimal.Decimal, 0)
		for tmpDateTime, tmpVal := range perValMap {
			yearLastVal, ok2 := yearLastValMap[tmpDateTime.Year()]
			if !ok2 {
				continue
			}

			tmpYearLastVal := decimal.NewFromFloat(yearLastVal)
			if tmpYearLastVal.IsZero() { //如果是0,那么就退出当前循环,进入下一个循环
				continue
			}
			divVal := decimal.NewFromFloat(tmpVal).Div(tmpYearLastVal)
			divValList = append(divValList, divVal)
		}

		lenDivValList := len(divValList)
		if lenDivValList < 2 {
			continue
		}

		divValSum := decimal.NewFromFloat(0)
		for _, divVal := range divValList {
			divValSum = divValSum.Add(divVal)
		}

		tmpDivVal := divValSum.Div(decimal.NewFromInt(int64(lenDivValList)))
		if tmpDivVal.IsZero() { //如果是0,那么就退出当前循环,进入下一个循环
			continue
		}
		currVal, _ := decimal.NewFromFloat(v.Value).Div(tmpDivVal).Round(4).Float64()

		dateDataMap[currDate] = currVal
	}

	return
}

func (obj BaseCalculate) Ljz() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	frequency := obj.Frequency         //需要变更的频度
	fromFrequency := obj.FromFrequency //来源的频度
	var isWeekData bool // 是否周度数据，如果是周度数据的话，是需要变频的，最后结果还需要除以7
	if fromFrequency == `周度` {
		isWeekData = true
	}

	fromEdbDataMap := make(map[time.Time]float64)
	if isWeekData {
		dataList, err = HandleDataByLinearRegressionByTime(dataList, fromEdbDataMap)
		if err != nil {
			return
		}
	}


	dateList := make([]time.Time, 0)
	valueMap := make(map[time.Time]float64)
	switch frequency {
	case "年度":
		yearMap := make(map[int]float64)
		yearList := make([]int, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			year := item.DataTime.Year()
			yearVal, ok := yearMap[year]
			if ok {
				yearMap[year] = item.Value + yearVal
			} else {
				yearList = append(yearList, year)
				yearMap[year] = item.Value
			}
		}
		for _, v := range yearList {
			currTime := time.Date(v, 12, 31, 0, 0, 0, 0, time.Local)
			dateList = append(dateList, currTime)
			valueMap[currTime] = yearMap[v]
		}
	case "半年度":
		yearMonthMap := make(map[string]float64)
		yearMonthList := make([]string, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			year := itemDate.Year()
			var tmpK string
			if itemDate.Month() <= 6 {
				tmpK = fmt.Sprint(year, "06")
			} else {
				tmpK = fmt.Sprint(year, "12")
			}

			yearVal, ok := yearMonthMap[tmpK]
			if ok {
				yearMonthMap[tmpK] = item.Value + yearVal
			} else {
				yearMonthList = append(yearMonthList, tmpK)
				yearMonthMap[tmpK] = item.Value
			}
		}
		for _, v := range yearMonthList {
			currTime, tmpErr := time.ParseInLocation(utils.FormatYearMonthUnSpace, v, time.Local)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			currTime = currTime.AddDate(0, 1, -1)
			dateList = append(dateList, currTime)
			valueMap[currTime] = yearMonthMap[v]
		}
	case "季度":
		yearMonthMap := make(map[string]float64)
		yearMonthList := make([]string, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			year := itemDate.Year()
			var tmpK string
			if itemDate.Month() <= 3 {
				tmpK = fmt.Sprint(year, "03")
			} else if itemDate.Month() <= 6 {
				tmpK = fmt.Sprint(year, "06")
			} else if itemDate.Month() <= 9 {
				tmpK = fmt.Sprint(year, "09")
			} else {
				tmpK = fmt.Sprint(year, "12")
			}

			yearVal, ok := yearMonthMap[tmpK]
			if ok {
				yearMonthMap[tmpK] = item.Value + yearVal
			} else {
				yearMonthList = append(yearMonthList, tmpK)
				yearMonthMap[tmpK] = item.Value
			}
		}
		for _, v := range yearMonthList {
			currTime, tmpErr := time.ParseInLocation(utils.FormatYearMonthUnSpace, v, time.Local)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			currTime = currTime.AddDate(0, 1, -1)
			dateList = append(dateList, currTime)
			valueMap[currTime] = yearMonthMap[v]
		}
	case "月度":
		yearMonthMap := make(map[string]float64)
		yearMonthList := make([]string, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			year := itemDate.Year()
			var tmpK string
			tmpK = fmt.Sprint(year*100 + int(itemDate.Month()))

			yearVal, ok := yearMonthMap[tmpK]
			if ok {
				yearMonthMap[tmpK] = item.Value + yearVal
			} else {
				yearMonthList = append(yearMonthList, tmpK)
				yearMonthMap[tmpK] = item.Value
			}
		}
		for _, v := range yearMonthList {
			currTime, tmpErr := time.ParseInLocation(utils.FormatYearMonthUnSpace, v, time.Local)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			currTime = currTime.AddDate(0, 1, -1)
			dateList = append(dateList, currTime)
			valueMap[currTime] = yearMonthMap[v]
		}
	case "旬度":
		tmpDateDataMap := make(map[time.Time]float64)
		tmpDateList := make([]time.Time, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			dayInt := itemDate.Year()*100 + int(itemDate.Month())
			var currTime time.Time
			if itemDate.Day() <= 10 {
				tmpK := fmt.Sprint(dayInt, "10")
				currTime, err = time.ParseInLocation(utils.FormatDateUnSpace, tmpK, time.Local)
				if err != nil {
					return
				}

			} else if itemDate.Day() <= 20 {
				tmpK := fmt.Sprint(dayInt, "20")
				currTime, err = time.ParseInLocation(utils.FormatDateUnSpace, tmpK, time.Local)
				if err != nil {
					return
				}
			} else {
				currTime, err = time.ParseInLocation(utils.FormatYearMonthUnSpace, fmt.Sprint(dayInt), time.Local)
				if err != nil {
					return
				}
				currTime = currTime.AddDate(0, 1, -1)
			}

			yearVal, ok := tmpDateDataMap[currTime]
			if ok {
				tmpDateDataMap[currTime] = item.Value + yearVal
			} else {
				tmpDateList = append(tmpDateList, currTime)
				tmpDateDataMap[currTime] = item.Value
			}
		}
		for _, currTime := range tmpDateList {
			dateList = append(dateList, currTime)
			valueMap[currTime] = tmpDateDataMap[currTime]
		}
	case "周度":
		tmpDateDataMap := make(map[time.Time]float64)
		tmpDateList := make([]time.Time, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			var currTime time.Time
			if itemDate.Weekday() == 0 {
				currTime = itemDate.AddDate(0, 0, 5)
			} else if itemDate.Weekday() == 6 {
				currTime = itemDate.AddDate(0, 0, 6)
			} else {
				currTime = itemDate.AddDate(0, 0, 5-int(itemDate.Weekday()))
			}

			yearVal, ok := tmpDateDataMap[currTime]
			if ok {
				tmpDateDataMap[currTime] = item.Value + yearVal
			} else {
				tmpDateList = append(tmpDateList, currTime)
				tmpDateDataMap[currTime] = item.Value
			}
		}
		for _, currTime := range tmpDateList {
			dateList = append(dateList, currTime)
			valueMap[currTime] = tmpDateDataMap[currTime]
		}
	default:
		err = errors.New("错误的频度")
		return

	}

	for _, currTime := range dateList {
		tmpVal, ok2 := valueMap[currTime]
		if !ok2 {
			err = errors.New("数据异常,date:" + currTime.Format(utils.FormatDate))
			return
		}
		var saveValue float64
		if isWeekData { //周度指标转的话，最后结果要除以7
			saveValue, _ = decimal.NewFromFloat(tmpVal).Div(decimal.NewFromInt(7)).Round(4).Float64()
		} else {
			saveValue, _ = decimal.NewFromFloat(tmpVal).Round(4).Float64()
		}

		dateDataMap[currTime] = saveValue
	}

	return
}

func (obj BaseCalculate) LjzNczj() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	frequency := obj.Frequency         //需要变更的频度
	fromFrequency := obj.FromFrequency //来源的频度

	var isWeekData bool // 是否周度数据，如果是周度数据的话，是需要变频的，最后结果还需要除以7
	if fromFrequency == `周度` {
		isWeekData = true
	}

	fromEdbDataMap := make(map[time.Time]float64)
	if isWeekData {
		dataList, err = HandleDataByLinearRegressionByTime(dataList, fromEdbDataMap)
		if err != nil {
			return
		}
	}


	dateList := make([]time.Time, 0)
	valueMap := make(map[time.Time]float64)
	yearMap := make(map[int]float64)
	switch frequency {
	case "周度":
		tmpDateDataMap := make(map[time.Time]float64)
		tmpDateList := make([]time.Time, 0)
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			var currTime time.Time
			if itemDate.Weekday() == 0 {
				currTime = itemDate.AddDate(0, 0, 5)
			} else if itemDate.Weekday() == 6 {
				currTime = itemDate.AddDate(0, 0, 6)
			} else {
				currTime = itemDate.AddDate(0, 0, 5-int(itemDate.Weekday()))
			}

			year := itemDate.Year()
			yearVal, ok := yearMap[year]
			if ok {
				yearMap[year] = item.Value + yearVal
			} else {
				yearMap[year] = item.Value
			}

			if itemDate.Equal(currTime) {
				tmpDateDataMap[itemDate] = yearMap[year]
				tmpDateList = append(tmpDateList, itemDate)
			}
		}
		for _, currTime := range tmpDateList {
			dateList = append(dateList, currTime)
			valueMap[currTime] = tmpDateDataMap[currTime]
		}
	default:
		for _, item := range dataList {
			item.DataTime = time.Date(item.DataTime.Year(), item.DataTime.Month(), item.DataTime.Day(), 0, 0, 0, 0, item.DataTime.Location())
			itemDate := item.DataTime
			year := itemDate.Year()
			yearVal, ok := yearMap[year]
			if ok {
				yearMap[year] = item.Value + yearVal
			} else {
				yearMap[year] = item.Value
			}
			valueMap[itemDate] = yearMap[year]
			dateList = append(dateList, itemDate)
		}

	}

	for _, currTime := range dateList {
		tmpVal, ok2 := valueMap[currTime]
		if !ok2 {
			err = errors.New("数据异常,date:" + currTime.Format(utils.FormatDate))
			return
		}
		var saveValue float64
		if isWeekData { //周度指标转的话，最后结果要除以7
			saveValue, _ = decimal.NewFromFloat(tmpVal).Div(decimal.NewFromInt(7)).Round(4).Float64()
		} else {
			saveValue, _ = decimal.NewFromFloat(tmpVal).Round(4).Float64()
		}

		dateDataMap[currTime] = saveValue
	}

	return
}

func (obj BaseCalculate) ExponentialSmoothing() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList
	formula := obj.Formula.(string) // alpha值

	alpha, _ := strconv.ParseFloat(formula, 64)
	if alpha <= 0 || alpha >= 1 {
		err = fmt.Errorf("alpha值有误: %v", alpha)
		return
	}

	fromDataList, err := calculateExponentialSmoothingData(dataList, alpha)
	if err != nil {
		return
	}

	for _, tmpData := range fromDataList {
		if tmpData.DataTime.IsZero() {
			continue
		}
		saveValue, _ := decimal.NewFromFloat(tmpData.Value).Round(4).Float64()

		dateDataMap[tmpData.DataTime] = saveValue
	}

	return
}

func calculateExponentialSmoothingData(dataList []*EdbInfoData, alpha float64) (newDataList []EdbInfoData, err error) {
	if alpha <= 0 || alpha >= 1 {
		err = fmt.Errorf("alpha值有误: %v", alpha)
		return
	}

	var preVal float64
	alphaDecimal := decimal.NewFromFloat(alpha)
	subAlpha := decimal.NewFromFloat(1).Sub(alphaDecimal)
	for k, d := range dataList {
		d.DataTime = time.Date(d.DataTime.Year(), d.DataTime.Month(), d.DataTime.Day(), 0, 0, 0, 0, d.DataTime.Location())
		if k == 0 {
			newDataList = append(newDataList, EdbInfoData{
				EdbDataId: k,
				DataTime:  dataList[k].DataTime,
				Value:     d.Value,
			})
			preVal = d.Value
			continue
		}

		preDecimal := decimal.NewFromFloat(preVal)
		valDecimal := decimal.NewFromFloat(d.Value)

		partA := alphaDecimal.Mul(valDecimal)
		partB := subAlpha.Mul(preDecimal)
		res, _ := (partA.Add(partB)).Float64()
		preVal = res
		newDataList = append(newDataList, EdbInfoData{
			EdbDataId: k,
			DataTime:  dataList[k].DataTime,
			Value:     res,
		})
	}
	return
}

func (obj BaseCalculate) Rjz() (dateDataMap map[time.Time]float64, err error, errMsg string) {
	dateDataMap = make(map[time.Time]float64)
	dataList := obj.DataList

	fromFrequency := obj.FromFrequency
	if fromFrequency == `` {
		errMsg = "原指标的频度不能为空"
		err = errors.New(errMsg)
		return
	}

	if obj.FromFrequency == "日度" {
		errMsg = "日度指标无需进行日均值计算"
		err = errors.New(errMsg)
		return
	}


	dataList = reverseSliceByDesc(dataList)

	var dateArr []time.Time
	dataMap := make(map[time.Time]*EdbInfoData)
	for _, v := range dataList {
		v.DataTime = time.Date(v.DataTime.Year(), v.DataTime.Month(), v.DataTime.Day(), 0, 0, 0, 0, v.DataTime.Location())
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
	}

	for _, currentDate := range dateArr {
		currentItem, ok := dataMap[currentDate]
		if !ok {
			continue
		}

		days := GetRjzFrequencyDays(currentDate, fromFrequency)
		val, ok := rjzDivV2(currentItem.Value, days)
		if !ok {
			continue
		}

		dateDataMap[currentDate] = val
	}

	return
}

func rjzDivV2(a float64, b int) (val float64, ok bool) {
	if b == 0 {
		return
	}

	af := decimal.NewFromFloat(a)
	bf := decimal.NewFromFloat(float64(b))
	val, _ = af.Div(bf).Round(4).Float64()
	ok = true

	return
}

func reverseSliceByDesc(slice []*EdbInfoData) (newSlice []*EdbInfoData) {
	if len(slice) <= 1 {
		newSlice = slice
		return
	}

	if !slice[1].DataTime.After(slice[0].DataTime) {
		newSlice = slice
		return
	}

	newSlice = make([]*EdbInfoData, len(slice))
	for i, j := 0, len(slice)-1; i < len(slice); i, j = i+1, j-1 {
		newSlice[i] = slice[j]
	}

	return
}

func EdbInfoSearchDataToData(dataList []*EdbInfoSearchData) (newDateDataList []*EdbInfoData, err error) {
	newDateDataList = make([]*EdbInfoData, 0)

	for _, data := range dataList {
		dateTime, tmpErr := time.ParseInLocation(utils.FormatDate, data.DataTime, time.Local)
		if tmpErr != nil {
			err = tmpErr
			return
		}

		newDateDataList = append(newDateDataList, &EdbInfoData{
			EdbDataId: data.EdbDataId,
			DataTime:  dateTime,
			Value:     data.Value,
		})
	}

	return
}

func GetDateDataAndDateList(dateDataMap map[time.Time]float64) (dateStrDataMap map[string]float64, dateStrList []string) {
	dateStrDataMap = make(map[string]float64)
	dateStrList = make([]string, 0)
	dateList := make([]time.Time, 0)
	for date, val := range dateDataMap {
		dateStr := date.Format(utils.FormatDate)
		dateStrDataMap[dateStr] = val
		dateList = append(dateList, date)
	}

	sort.Slice(dateList, func(i, j int) bool {
		return dateList[i].Before(dateList[j])
	})

	for _, v := range dateList {
		dateStrList = append(dateStrList, v.Format(utils.FormatDate))
	}

	return
}

type StepCalculateBySearchData struct {
	DataList   []*EdbInfoSearchData `description:"基础数据"`
	Calculates []CalculatesReq
}

type CalculatesReq struct {
	Formula       interface{}
	Calendar      string `description:"公历/农历"`
	Frequency     string `description:"需要转换的频度"`
	MoveType      int    `description:"移动方式:1:领先(默认),2:滞后"`
	MoveFrequency string `description:"移动频度"`
	FromFrequency string `description:"来源的频度"`
	Source        int    `description:"1：累计值转月；2：累计值转季；3：同比值；4：同差值；5：N数值移动平均数计算；6：环比值；7：环差值；8：升频；9：降频；10：时间移位；11：超季节性；12：年化；13：累计值；14：累计值年初至今；15：指数修匀；16：日均值"`
	Sort          int    `description:"计算顺序"`
}

func TransDateData2EdbData(dateData map[time.Time]float64) (edbData []*EdbInfoData) {
	edbData = make([]*EdbInfoData, 0)
	for d, v := range dateData {
		edbData = append(edbData, &EdbInfoData{
			DataTime: d,
			Value:    v,
		})
	}
	sort.Slice(edbData, func(i, j int) bool {
		return edbData[i].DataTime.Before(edbData[j].DataTime)
	})
	return
}

func EdbInfoSearchDataToEdbInfoData(list []*EdbInfoSearchData) (result []*EdbInfoData, err error) {
	result = make([]*EdbInfoData, 0)
	for _, v := range list {
		dataTime, tmpErr := time.ParseInLocation(utils.FormatDate, v.DataTime, time.Local)
		if tmpErr != nil {
			err = tmpErr
			return
		}
		result = append(result, &EdbInfoData{
			EdbDataId: v.EdbDataId,
			DataTime:  dataTime,
			Value:     v.Value,
		})
	}

	return
}
