package models

import (
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/models/mgo"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"github.com/shopspring/decimal"
	"go.mongodb.org/mongo-driver/bson"
	"strconv"
	"strings"
	"time"
)

type ManualEdbdata struct {
	TradeCode  string    `gorm:"column:TRADE_CODE;primaryKey" description:"指标编码"`
	Dt         string    `gorm:"column:DT" description:"日期"`
	Close      string    `gorm:"column:CLOSE" description:"值"`
	ModifyTime time.Time `gorm:"column:modify_time" description:"修改时间"`
}

func (m *ManualEdbdata) TableName() string {
	return "edb_data_manual"
}

func GetEdbdataManualByCondition(condition string, pars []interface{}) (item []*ManualEdbdata, err error) {
	sql := ` SELECT  * FROM edbdata WHERE 1=1  `
	if condition != "" {
		sql += condition
	}
	sql += ` ORDER BY DT DESC `
	err = global.DmSQL["edb"].Raw(sql, pars...).Find(&item).Error
	return
}

func AddEdbDataFromManual(edbCode string) (err error) {
	var condition string
	var pars []interface{}

	if edbCode != "" {
		condition += " AND TRADE_CODE=? "
		pars = append(pars, edbCode)
	}

	manualDataList, err := GetEdbdataManualByCondition(condition, pars)
	if err != nil {
		return
	}
	dataLen := len(manualDataList)

	if dataLen > 0 {
		var isAdd bool
		addSql := ` INSERT INTO edb_data_manual(edb_info_id,edb_code,data_time,value,create_time,modify_time,data_timestamp) values `
		for i := 0; i < dataLen; i++ {
			item := manualDataList[i]
			eDate := item.Dt
			sValue := item.Close
			tmpDecimal, err := decimal.NewFromString(sValue)
			if err != nil {
				return err
			}
			sValue = tmpDecimal.Round(4).String()

			dataTime, err := time.ParseInLocation(utils.FormatDate, eDate, time.Local)
			if err != nil {
				return err
			}
			timestamp := dataTime.UnixNano() / 1e6
			timeStr := fmt.Sprintf("%d", timestamp)
			addSql += GetAddSql("0", edbCode, eDate, timeStr, sValue)
			isAdd = true
		}
		if isAdd {
			addSql = strings.TrimRight(addSql, ",")
			err = global.DEFAULT_DmSQL.Exec(addSql).Error
			if err != nil {
				return
			}
		}
	}
	return
}

func RefreshEdbDataFromManual(edbInfoId int, edbCode, startDate string) (err error) {
	source := utils.DATA_SOURCE_MANUAL
	subSource := utils.DATA_SUB_SOURCE_EDB

	if err != nil {
		return
	}
	edbInfoIdStr := strconv.Itoa(edbInfoId)
	var condition string
	var pars []interface{}

	if edbCode != "" {
		condition += " AND TRADE_CODE=? "
		pars = append(pars, edbCode)
	}

	if startDate != "" {
		condition += " AND DT>=? "
		pars = append(pars, startDate)
	} else {
		condition += " AND DT != ? "
		pars = append(pars, `0000-00-00`)
	}

	manualDataList, err := GetEdbdataManualByCondition(condition, pars)
	if err != nil {
		return
	}

	var realDataMaxDate, edbDataInsertConfigDate time.Time
	var edbDataInsertConfig *EdbDataInsertConfig
	var isFindConfigDateRealData bool //是否找到配置日期的实际数据的值
	edbDataInsertConfigDateStr := ``
	{
		edbDataInsertConfig, err = GetEdbDataInsertConfigByEdbId(edbInfoId)
		if err != nil && err.Error() != utils.ErrNoRow() {
			return
		}
		if edbDataInsertConfig != nil {
			edbDataInsertConfigDate = edbDataInsertConfig.Date
			edbDataInsertConfigDateStr = edbDataInsertConfig.Date.Format(utils.FormatDate)
		}
	}

	var existCondition string
	var existPars []interface{}

	existCondition += " AND edb_info_id=? "
	existPars = append(existPars, edbInfoId)
	if startDate != "" {
		existCondition += " AND data_time>=? "
		existPars = append(existPars, startDate)
	}

	existList, err := GetEdbDataByCondition(source, subSource, existCondition, existPars)
	if err != nil {
		return err
	}
	existMap := make(map[string]*EdbInfoSearchData)
	for _, v := range existList {
		existMap[v.DataTime] = v
	}

	addSql := ` INSERT INTO edb_data_manual (edb_info_id,edb_code,data_time,value,create_time,modify_time,data_timestamp) values `
	var isAdd bool
	manualMap := make(map[string]*ManualEdbdata)
	for _, v := range manualDataList {
		item := v
		eDate := item.Dt
		sValue := item.Close
		tmpDecimal, err := decimal.NewFromString(sValue)
		if err != nil {
			return err
		}
		sValue = tmpDecimal.Round(4).String()

		dataTime, err := time.ParseInLocation(utils.FormatDate, eDate, time.Local)
		if err != nil {
			return err
		}

		if findItem, ok := existMap[v.Dt]; !ok {
			timestamp := dataTime.UnixNano() / 1e6
			timeStr := fmt.Sprintf("%d", timestamp)

			addSql += GetAddSql(edbInfoIdStr, edbCode, eDate, timeStr, sValue)
			isAdd = true
		} else {
			if findItem != nil && utils.SubFloatToString(findItem.Value, 30) != item.Close {
				err = ModifyEdbDataById(source, subSource, findItem.EdbDataId, item.Close)
				if err != nil {
					return err
				}
			}
		}
		manualMap[v.Dt] = v

		{
			if realDataMaxDate.IsZero() || dataTime.After(realDataMaxDate) {
				realDataMaxDate = dataTime
			}
			if edbDataInsertConfigDate.IsZero() || dataTime.Equal(edbDataInsertConfigDate) {
				isFindConfigDateRealData = true
			}
		}
	}
	for _, v := range existList {
		if _, ok := manualMap[v.DataTime]; !ok {

			if edbDataInsertConfigDateStr == `` || edbDataInsertConfigDateStr != v.DataTime {
				DeleteEdbDataById(utils.DATA_SOURCE_MANUAL, subSource, v.EdbDataId)
			}
		}
	}
	if isAdd {
		addSql = strings.TrimRight(addSql, ",")
		err = global.DEFAULT_DmSQL.Exec(addSql).Error
		if err != nil {
			fmt.Println("RefreshAllEdbDataByManual add Err", err.Error())
			return
		}
	}

	HandleConfigInsertEdbData(realDataMaxDate, edbDataInsertConfig, edbInfoId, source, subSource, existMap, isFindConfigDateRealData)

	return
}

func HandleConfigInsertEdbData(realDataMaxDate time.Time, edbDataInsertConfig *EdbDataInsertConfig, edbInfoId, source, subSource int, existMap map[string]*EdbInfoSearchData, isFindConfigDateRealData bool) {
	if edbDataInsertConfig == nil {
		return
	}
	var err error
	defer func() {
		if err != nil {
			fmt.Println("处理手工数据补充的配置失败,err:", err)
		}
	}()
	edbDataInsertConfigDate := edbDataInsertConfig.Date // 配置的日期

	if realDataMaxDate.After(edbDataInsertConfigDate) || realDataMaxDate.Equal(edbDataInsertConfigDate) {
		go DeleteEdbDataInsertConfigByEdbId(edbInfoId)

		edbDataInsertConfigDateStr := edbDataInsertConfigDate.Format(utils.FormatDate)
		if item, ok := existMap[edbDataInsertConfigDateStr]; ok && !isFindConfigDateRealData {
			DeleteEdbDataById(source, subSource, item.EdbDataId)
		}
	} else {
		edbDataInsertConfig.RealDate = realDataMaxDate
		err = global.DEFAULT_DmSQL.Model(&EdbDataInsertConfig{}).Where("edb_info_id=?", edbInfoId).Update("RealDate", realDataMaxDate).Error
	}

	return
}

func HandleConfigInsertEdbDataByMongo(realDataMaxDate time.Time, edbDataInsertConfig *EdbDataInsertConfig, edbInfoId, source, subSource int, existMap map[string]*mgo.EdbDataBusiness, isFindConfigDateRealData bool) {
	if edbDataInsertConfig == nil {
		return
	}
	var err error
	defer func() {
		if err != nil {
			fmt.Println("处理手工数据补充的配置失败,err:", err)
		}
	}()

	edbDataInsertConfigDate := edbDataInsertConfig.Date // 配置的日期

	if realDataMaxDate.After(edbDataInsertConfigDate) || realDataMaxDate.Equal(edbDataInsertConfigDate) {
		go DeleteEdbDataInsertConfigByEdbId(edbInfoId)

		mogDataObj := mgo.EdbDataBusiness{}
		coll := mogDataObj.GetCollection()
		edbDataInsertConfigDateStr := edbDataInsertConfigDate.Format(utils.FormatDate)
		if item, ok := existMap[edbDataInsertConfigDateStr]; ok && !isFindConfigDateRealData {
			mogDataObj.RemoveManyByColl(coll, bson.M{"_id": item.ID})
		}
	} else {
		edbDataInsertConfig.RealDate = realDataMaxDate
		err = global.DEFAULT_DmSQL.Model(&EdbDataInsertConfig{}).Where("edb_info_id=?", edbInfoId).Update("RealDate", realDataMaxDate).Error
	}

	return
}
