package models

import (
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"fmt"

	"github.com/shopspring/decimal"
	"gorm.io/gorm"
	"sort"
	"strings"
	"time"
)

type EdbDataCalculateLjztbpj struct {
	EdbDataId     int       `gorm:"primaryKey;autoIncrement;column:edb_data_id"`
	EdbInfoId     int       `gorm:"column:edb_info_id" description:"指标ID"`
	EdbCode       string    `gorm:"column:edb_code" description:"指标编码"`
	DataTime      string    `gorm:"column:data_time" description:"数据日期"`
	Value         float64   `gorm:"column:value" description:"数据值"`
	Status        int       `gorm:"column:status" description:"状态"`
	CreateTime    time.Time `gorm:"column:status" description:"创建时间"`
	ModifyTime    time.Time `gorm:"column:create_time" description:"修改时间"`
	DataTimestamp int64     `gorm:"column:modify_time" description:"数据日期时间戳"`
}

func AddCalculateLjztbpj(req *EdbInfoCalculateBatchSaveReq, firstEdbInfo, secondEdbInfo *EdbInfo, edbCode, uniqueCode string, sysUserId int, sysUserRealName string) (edbInfo *EdbInfo, err error) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()
	pjDataMap := make(map[string]float64)

	pjEdbDataTimeList := make([]string, 0)

	var lastDateTime time.Time
	{
		var condition string
		var pars []interface{}

		condition += " AND data_time like ? AND edb_info_id=? "
		pars = append(pars, "%12-31", firstEdbInfo.EdbInfoId)

		lastEdbData, tmpErr := GetLastEdbData(condition, pars, firstEdbInfo.Source, firstEdbInfo.SubSource)
		if tmpErr != nil {
			err = tmpErr
			return
		}
		lastDateTime, _ = time.ParseInLocation(utils.FormatDate, lastEdbData.DataTime, time.Local)

		firstDataList, tmpErr := GetEdbDataListAllByTo(to, firstEdbInfo.Source, firstEdbInfo.SubSource, FindEdbDataListAllCond{
			EdbInfoId:       firstEdbInfo.EdbInfoId,
			EndDataTime:     lastEdbData.DataTime,
			EndDataTimeCond: "<=",
		}, 0)
		if tmpErr != nil {
			err = tmpErr
			return
		}

		for _, v := range firstDataList {
			pjDataMap[v.DataTime] = v.Value
			pjEdbDataTimeList = append(pjEdbDataTimeList, v.DataTime)
		}
	}

	if req.EdbInfoId <= 0 {
		edbInfo = &EdbInfo{
			SourceName:       "累计值同比拼接",
			Source:           utils.DATA_SOURCE_CALCULATE_LJZTBPJ,
			EdbCode:          edbCode,
			EdbName:          req.EdbName,
			EdbNameSource:    req.EdbName,
			Frequency:        req.Frequency,
			Unit:             req.Unit,
			StartDate:        firstEdbInfo.StartDate,
			EndDate:          firstEdbInfo.EndDate,
			ClassifyId:       req.ClassifyId,
			SysUserId:        sysUserId,
			SysUserRealName:  sysUserRealName,
			UniqueCode:       uniqueCode,
			CreateTime:       time.Now(),
			ModifyTime:       time.Now(),
			CalculateFormula: lastDateTime.Format(utils.FormatDate),
			EdbNameEn:        req.EdbName,
			UnitEn:           req.Unit,
			EdbType:          2,
			Sort:             GetAddEdbMaxSortByClassifyId(req.ClassifyId, utils.EDB_INFO_TYPE),
		}
		tmpErr := to.Create(edbInfo).Error
		if tmpErr != nil {
			err = tmpErr
			return
		}

	} else {
		edbInfo, err = GetEdbInfoById(req.EdbInfoId)
		if err != nil {
			return
		}
		tmpEdbInfo, tmpErr := GetEdbInfoById(edbInfo.EdbInfoId)
		if tmpErr != nil {
			err = tmpErr
			return
		}
		tmpEdbInfo.EdbName = req.EdbName
		tmpEdbInfo.ClassifyId = req.ClassifyId
		tmpEdbInfo.Frequency = req.Frequency
		tmpEdbInfo.Unit = req.Unit
		tmpEdbInfo.CalculateFormula = req.Formula

		edbInfo = tmpEdbInfo

		dataTableName := GetEdbDataTableName(utils.DATA_SOURCE_CALCULATE_LJZTBPJ, utils.DATA_SUB_SOURCE_EDB)
		fmt.Println("dataTableName:" + dataTableName)
		deleteSql := ` DELETE FROM %s WHERE edb_info_id=? `
		deleteSql = fmt.Sprintf(deleteSql, dataTableName)
		err = to.Exec(deleteSql, req.EdbInfoId).Error
		if err != nil {
			return
		}

		sql := ` DELETE FROM edb_info_calculate_mapping WHERE edb_info_id=? `
		err = to.Exec(sql, edbInfo.EdbInfoId).Error
	}


	var existItemA, existItemB *EdbInfoCalculateMapping
	{
		existItemA = &EdbInfoCalculateMapping{
			EdbInfoCalculateMappingId: 0,
			EdbInfoId:                 edbInfo.EdbInfoId,
			Source:                    edbInfo.Source,
			SourceName:                edbInfo.SourceName,
			EdbCode:                   edbInfo.EdbCode,
			FromEdbInfoId:             firstEdbInfo.EdbInfoId,
			FromEdbCode:               firstEdbInfo.EdbCode,
			FromEdbName:               firstEdbInfo.EdbName,
			FromSource:                firstEdbInfo.Source,
			FromSourceName:            firstEdbInfo.SourceName,
			FromTag:                   "A",
			Sort:                      1,
			CreateTime:                time.Now(),
			ModifyTime:                time.Now(),
			FromSubSource:             firstEdbInfo.SubSource,
		}
		tmpErr := to.Create(existItemA).Error
		if tmpErr != nil {
			err = tmpErr
			return
		}

	}

	{
		existItemB = &EdbInfoCalculateMapping{
			EdbInfoCalculateMappingId: 0,
			EdbInfoId:                 edbInfo.EdbInfoId,
			Source:                    edbInfo.Source,
			SourceName:                edbInfo.SourceName,
			EdbCode:                   edbInfo.EdbCode,
			FromEdbInfoId:             secondEdbInfo.EdbInfoId,
			FromEdbCode:               secondEdbInfo.EdbCode,
			FromEdbName:               secondEdbInfo.EdbName,
			FromSource:                secondEdbInfo.Source,
			FromSourceName:            secondEdbInfo.SourceName,
			FromTag:                   "B",
			Sort:                      1,
			CreateTime:                time.Now(),
			ModifyTime:                time.Now(),
			FromSubSource:             secondEdbInfo.SubSource,
		}
		tmpErr := to.Create(existItemB).Error
		if tmpErr != nil {
			err = tmpErr
			return
		}

	}

	err = refreshAllCalculateLjztbpj(to, edbInfo, existItemA, existItemB)

	return
}

func EditCalculateLjztbpj(req *EdbInfoCalculateBatchEditReq, edbInfo, firstEdbInfo, secondEdbInfo *EdbInfo) (err error) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	nowEdbInfo := *edbInfo // 现在的指标信息

	sql := ``
	edbInfo.EdbNameSource = req.EdbName
	edbInfo.Frequency = req.Frequency
	edbInfo.Unit = req.Unit
	edbInfo.ClassifyId = req.ClassifyId
	edbInfo.EdbNameEn = req.EdbNameEn
	edbInfo.UnitEn = req.UnitEn
	edbInfo.ModifyTime = time.Now()
	err = to.Model(edbInfo).Select([]string{"EdbName", "EdbNameSource", "Frequency", "Unit", "ClassifyId", "ModifyTime", "EdbNameEn", "UnitEn"}).Updates(edbInfo).Error
	if err != nil {
		return
	}

	var existCondition string
	var existPars []interface{}
	existCondition += " AND edb_info_id=? "
	existPars = append(existPars, edbInfo.EdbInfoId)
	existList, err := GetEdbInfoCalculateListByCondition(existCondition, existPars)
	if err != nil {
		err = errors.New("判断指标是否改变失败,Err:" + err.Error())
		return
	}

	var existItemA, existItemB *EdbInfoCalculateMapping
	for _, existItem := range existList {
		if existItem.FromTag == "A" {
			existItemA = existItem
		} else if existItem.FromTag == "B" {
			existItemB = existItem
		}
	}

	isNeedCalculateData := false

	if req.Formula != nowEdbInfo.CalculateFormula {
		isNeedCalculateData = true
	}

	if existItemA.FromEdbInfoId != firstEdbInfo.EdbInfoId {
		sql := ` DELETE FROM edb_info_calculate_mapping WHERE edb_info_id = ? and from_edb_info_id = ?`
		err = to.Exec(sql, edbInfo.EdbInfoId, existItemA.FromEdbInfoId).Error
		if err != nil {
			err = errors.New("删除拼接日期之前的指标关联关系失败,Err:" + err.Error())
			return
		}

		{
			existItemA = &EdbInfoCalculateMapping{
				EdbInfoCalculateMappingId: 0,
				EdbInfoId:                 edbInfo.EdbInfoId,
				Source:                    edbInfo.Source,
				SourceName:                edbInfo.SourceName,
				EdbCode:                   edbInfo.EdbCode,
				FromEdbInfoId:             firstEdbInfo.EdbInfoId,
				FromEdbCode:               firstEdbInfo.EdbCode,
				FromEdbName:               firstEdbInfo.EdbName,
				FromSource:                firstEdbInfo.Source,
				FromSourceName:            firstEdbInfo.SourceName,
				FromTag:                   "A",
				Sort:                      1,
				CreateTime:                time.Now(),
				ModifyTime:                time.Now(),
				FromSubSource:             firstEdbInfo.SubSource,
			}
			tmpErr := to.Create(existItemA).Error
			if tmpErr != nil {
				err = tmpErr
				return
			}

			isNeedCalculateData = true
		}
	}

	if existItemB.FromEdbInfoId != secondEdbInfo.EdbInfoId {
		sql = ` DELETE FROM edb_info_calculate_mapping WHERE edb_info_id = ? and from_edb_info_id = ?`
		err = to.Exec(sql, edbInfo.EdbInfoId, existItemB.FromEdbInfoId).Error
		if err != nil {
			err = errors.New("删除拼接日期之后的指标关联关系失败,Err:" + err.Error())
			return
		}

		{
			existItemB = &EdbInfoCalculateMapping{
				EdbInfoCalculateMappingId: 0,
				EdbInfoId:                 edbInfo.EdbInfoId,
				Source:                    edbInfo.Source,
				SourceName:                edbInfo.SourceName,
				EdbCode:                   edbInfo.EdbCode,
				FromEdbInfoId:             secondEdbInfo.EdbInfoId,
				FromEdbCode:               secondEdbInfo.EdbCode,
				FromEdbName:               secondEdbInfo.EdbName,
				FromSource:                secondEdbInfo.Source,
				FromSourceName:            secondEdbInfo.SourceName,
				FromTag:                   "B",
				Sort:                      2,
				CreateTime:                time.Now(),
				ModifyTime:                time.Now(),
				FromSubSource:             secondEdbInfo.SubSource,
			}

			tmpErr := to.Create(existItemB).Error
			if tmpErr != nil {
				err = tmpErr
				return
			}

			isNeedCalculateData = true
		}
	}

	if isNeedCalculateData {
		tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
		sql = fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? `, tableName)
		err = to.Exec(sql, edbInfo.EdbInfoId).Error
		if err != nil {
			err = errors.New("删除所有的累计值同比拼接指标数据失败,Err:" + err.Error())
			return
		}

		{
			var condition string
			var pars []interface{}

			condition += " AND data_time like ? AND edb_info_id=? "
			pars = append(pars, "%12-31", firstEdbInfo.EdbInfoId)

			lastEdbData, tmpErr := GetLastEdbData(condition, pars, firstEdbInfo.Source, firstEdbInfo.SubSource)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			lastDateTime, tmpErr := time.ParseInLocation(utils.FormatDate, lastEdbData.DataTime, time.Local)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			edbInfo.CalculateFormula = lastDateTime.Format(utils.FormatDate)
			err = to.Model(edbInfo).Select([]string{"CalculateFormula"}).Updates(edbInfo).Error
			if err != nil {
				return
			}

		}
		err = refreshAllCalculateLjztbpj(to, edbInfo, existItemA, existItemB)
	}

	return
}

func RefreshAllCalculateLjztbpj(edbInfo *EdbInfo) (err error) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	var existCondition string
	var existPars []interface{}
	existCondition += " AND edb_info_id=? "
	existPars = append(existPars, edbInfo.EdbInfoId)
	existList, err := GetEdbInfoCalculateListByCondition(existCondition, existPars)
	if err != nil {
		err = errors.New("判断指标是否改变失败,Err:" + err.Error())
		return
	}

	var existItemA, existItemB *EdbInfoCalculateMapping
	for _, existItem := range existList {
		if existItem.FromTag == "A" {
			existItemA = existItem
		} else if existItem.FromTag == "B" {
			existItemB = existItem
		}
	}
	err = refreshAllCalculateLjztbpj(to, edbInfo, existItemA, existItemB)
	return
}

func refreshAllCalculateLjztbpj(to *gorm.DB, edbInfo *EdbInfo, existItemA, existItemB *EdbInfoCalculateMapping) (err error) {
	var condition string
	var pars []interface{}
	condition += " AND edb_info_id=? "
	pars = append(pars, edbInfo.EdbInfoId)

	var dataList []*EdbDataCalculateLjztbpj
	sql := ` SELECT * FROM edb_data_calculate_ljztbpj WHERE edb_info_id=? ORDER BY data_time DESC `
	err = to.Raw(sql, edbInfo.EdbInfoId).Scan(&dataList).Error
	if err != nil {
		return err
	}

	pjDataMap := make(map[string]float64)     //需要入库的数据
	nowEdbDataMap := make(map[string]float64) //当前指标的数据（已经在库里了，不需要重新）
	pjEdbDataTimeList := make([]string, 0)

	dataMap := make(map[string]*EdbDataCalculateLjztbpj)
	for _, v := range dataList {
		pjEdbDataTimeList = append(pjEdbDataTimeList, v.DataTime)
		dataMap[v.DataTime] = v
		nowEdbDataMap[v.DataTime] = v.Value
	}

	startCalculationDate, _ := time.ParseInLocation(utils.FormatDate, edbInfo.CalculateFormula, time.Local)

	{
		firstDataList, tmpErr := GetEdbDataListAllByTo(to, existItemA.FromSource, existItemA.FromSubSource, FindEdbDataListAllCond{
			EdbInfoId:       existItemA.FromEdbInfoId,
			EndDataTime:     startCalculationDate.Format(utils.FormatDate),
			EndDataTimeCond: "<=",
		}, 0)
		if tmpErr != nil {
			return tmpErr
		}

		for _, v := range firstDataList {
			if edbData, ok := dataMap[v.DataTime]; ok {
				if edbData.Value != v.Value {
					edbData.Value = v.Value
					_ = to.Model(edbData).Select([]string{"Value"}).Updates(edbData).Error
				}
			} else {
				pjDataMap[v.DataTime] = v.Value
				pjEdbDataTimeList = append(pjEdbDataTimeList, v.DataTime)
			}

			nowEdbDataMap[v.DataTime] = v.Value
		}
	}

	tbzEdbDataMap := make(map[string]float64)

	tbzEdbDataTimeList := make([]string, 0)

	{
		secondDataList, tmpErr := GetEdbDataListAllByTo(to, existItemB.FromSource, existItemB.FromSubSource, FindEdbDataListAllCond{
			EdbInfoId:         existItemB.FromEdbInfoId,
			StartDataTime:     startCalculationDate.Format(utils.FormatDate),
			StartDataTimeCond: ">",
		}, 0)
		if tmpErr != nil {
			return tmpErr
		}

		for _, v := range secondDataList {
			tbzEdbDataMap[v.DataTime] = v.Value
			tbzEdbDataTimeList = append(tbzEdbDataTimeList, v.DataTime)
		}
	}

	sort.Strings(tbzEdbDataTimeList)

	removeDateList := make([]string, 0)
	for nowEdbDate := range nowEdbDataMap {
		nowEdbDateTime, _ := time.ParseInLocation(utils.FormatDate, nowEdbDate, time.Local)
		if startCalculationDate.Before(nowEdbDateTime) {
			if _, ok := tbzEdbDataMap[nowEdbDate]; !ok {
				removeDateList = append(removeDateList, nowEdbDate)
			}
		}
	}

	updateEdbDataMap := make(map[string]float64)
	for _, v := range tbzEdbDataTimeList {
		tbzDataTime, _ := time.ParseInLocation(utils.FormatDate, v, time.Local)

		var pjDataTime time.Time
		if tbzDataTime.Month() == 2 {
			pjDataTime = tbzDataTime.AddDate(0, -11, 0)
			pjDataTime = time.Date(pjDataTime.Year(), pjDataTime.Month(), 1, 0, 0, 0, 0, time.Now().Location()).AddDate(0, 0, -1)
		} else {
			pjDataTime = tbzDataTime.AddDate(-1, 0, 0)
		}

		if nowEdbDataValue, isHas := nowEdbDataMap[v]; isHas {
			if lastYearEdbDataValue, ok := nowEdbDataMap[pjDataTime.Format(utils.FormatDate)]; ok {

				if v == "2021-08-31" {
					fmt.Println("进来了")
				}
				tbzDataValue := tbzEdbDataMap[v] //同比值
				currValue := lastYearEdbDataValue * (1 + tbzDataValue/100)
				currValue, _ = decimal.NewFromFloat(currValue).Truncate(4).Float64() //保留4位小数
				if nowEdbDataValue != currValue {
					updateEdbDataMap[v] = currValue
				}
			}
		} else {

			if pjDataValue, ok := nowEdbDataMap[pjDataTime.Format(utils.FormatDate)]; ok {
				tbzDataValue := tbzEdbDataMap[v] //同比值
				currValue := pjDataValue * (1 + tbzDataValue/100)

				currValue, _ = decimal.NewFromFloat(currValue).Truncate(4).Float64()
				nowEdbDataMap[v] = currValue

				pjDataMap[v] = currValue
				pjEdbDataTimeList = append(pjEdbDataTimeList, v)
			}
		}
	}

	{
		addDataList := make([]*EdbDataCalculateLjztbpj, 0)
		sort.Strings(pjEdbDataTimeList)
		for _, dataTime := range pjEdbDataTimeList {
			if dataValue, ok := pjDataMap[dataTime]; ok {
				currentDate, _ := time.ParseInLocation(utils.FormatDate, dataTime, time.Local)
				timestamp := currentDate.UnixNano() / 1e6

				edbDataLjztbpj := &EdbDataCalculateLjztbpj{
					EdbInfoId:     edbInfo.EdbInfoId,
					EdbCode:       edbInfo.EdbCode,
					DataTime:      dataTime,
					Value:         dataValue,
					Status:        1,
					CreateTime:    time.Now(),
					ModifyTime:    time.Now(),
					DataTimestamp: timestamp,
				}
				addDataList = append(addDataList, edbDataLjztbpj)
			}
		}

		if len(addDataList) > 0 {
			tmpErr := to.CreateInBatches(addDataList, 500).Error
			if tmpErr != nil {
				err = tmpErr
				return
			}
		}
	}

	{
		if len(removeDateList) > 0 {
			removeDateStr := strings.Join(removeDateList, `","`)
			removeDateStr = `"` + removeDateStr + `"`
			tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
			sql := fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? and data_time in (%s) `, tableName, removeDateStr)

			err = to.Exec(sql, edbInfo.EdbInfoId).Error
			if err != nil {
				err = errors.New("删除不存在的累计值同比拼接指标数据失败,Err:" + err.Error())
				return
			}
		}
	}

	{
		tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
		for edbDate, edbDataValue := range updateEdbDataMap {
			sql := fmt.Sprintf(` UPDATE  %s set value = ?,modify_time=now() WHERE edb_info_id = ? and data_time = ? `, tableName)

			err = to.Exec(sql, edbDataValue, edbInfo.EdbInfoId, edbDate).Error
			if err != nil {
				err = errors.New("更新现有的累计值同比拼接指标数据失败,Err:" + err.Error())
				return
			}
		}
	}
	return
}
