package models

import (
	"encoding/json"
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"github.com/shopspring/decimal"
	"gorm.io/gorm"
	"reflect"
	"strconv"
	"strings"
	"time"
)

type Percentile struct {
}

func (obj Percentile) Add(params AddCalculateBatchParams) (edbInfo *EdbInfo, err error, errMsg string) {
	req := params.Req
	fromEdbInfo := params.FromEdbInfo
	edbCode := params.EdbCode
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	edbInfo = new(EdbInfo)
	edbInfo.Source = obj.GetSource()
	edbInfo.SourceName = obj.GetSourceName()
	edbInfo.EdbCode = edbCode
	edbInfo.EdbName = req.EdbName
	edbInfo.EdbNameSource = req.EdbName
	edbInfo.Frequency = req.Frequency
	edbInfo.Unit = req.Unit
	edbInfo.ClassifyId = req.ClassifyId
	edbInfo.SysUserId = params.SysUserId
	edbInfo.SysUserRealName = params.SysUserRealName
	edbInfo.CreateTime = time.Now()
	edbInfo.ModifyTime = time.Now()
	edbInfo.UniqueCode = params.UniqueCode
	edbInfo.CalculateFormula = req.Formula
	edbInfo.EdbNameEn = req.EdbName
	edbInfo.UnitEn = req.Unit
	edbInfo.EdbType = obj.GetEdbType()
	edbInfo.Sort = GetAddEdbMaxSortByClassifyId(req.ClassifyId, utils.EDB_INFO_TYPE)
	tmpErr := to.Create(edbInfo).Error
	if tmpErr != nil {
		err = tmpErr
		return
	}

	{
		calculateMappingItem := new(EdbInfoCalculateMapping)
		calculateMappingItem.CreateTime = time.Now()
		calculateMappingItem.ModifyTime = time.Now()
		calculateMappingItem.Sort = 1
		calculateMappingItem.EdbCode = edbCode
		calculateMappingItem.EdbInfoId = edbInfo.EdbInfoId
		calculateMappingItem.FromEdbInfoId = fromEdbInfo.EdbInfoId
		calculateMappingItem.FromEdbCode = fromEdbInfo.EdbCode
		calculateMappingItem.FromEdbName = fromEdbInfo.EdbName
		calculateMappingItem.FromSource = fromEdbInfo.Source
		calculateMappingItem.FromSourceName = fromEdbInfo.SourceName
		calculateMappingItem.FromTag = ""
		calculateMappingItem.Source = edbInfo.Source
		calculateMappingItem.SourceName = edbInfo.SourceName
		calculateMappingItem.FromSubSource = fromEdbInfo.SubSource
		err = to.Create(calculateMappingItem).Error
		if err != nil {
			return
		}
	}

	err, errMsg = obj.refresh(to, edbInfo, fromEdbInfo, edbInfo.EdbCode)

	return
}

func (obj Percentile) Edit(params EditCalculateBatchParams) (err error, errMsg string) {
	req := params.Req
	edbInfo := params.EdbInfo
	fromEdbInfo := params.FromEdbInfo

	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()
	tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)

	oldEdbInfo := *edbInfo

	edbInfo.EdbName = req.EdbName
	edbInfo.EdbNameSource = req.EdbName
	edbInfo.Frequency = req.Frequency
	edbInfo.Unit = req.Unit
	edbInfo.ClassifyId = req.ClassifyId
	edbInfo.CalculateFormula = req.Formula
	edbInfo.EdbNameEn = req.EdbNameEn
	edbInfo.UnitEn = req.UnitEn
	edbInfo.ModifyTime = time.Now()
	err = to.Model(edbInfo).Select([]string{"EdbName", "EdbNameSource", "Frequency", "Unit", "ClassifyId", "CalculateFormula", "ModifyTime", "EdbNameEn", "UnitEn"}).Updates(edbInfo).Error
	if err != nil {
		return
	}

	var existCondition string
	var existPars []interface{}
	existCondition += " AND edb_info_id=? AND from_edb_info_id=? "
	existPars = append(existPars, edbInfo.EdbInfoId, req.FromEdbInfoId)

	count, err := GetEdbInfoCalculateCountByCondition(existCondition, existPars)
	if err != nil {
		err = errors.New("判断指标是否改变失败,Err:" + err.Error())
		return
	}
	if count > 0 { // 指标未被替换，无需处理逻辑
		if oldEdbInfo.CalculateFormula != edbInfo.CalculateFormula {
			err, errMsg = obj.refresh(to, edbInfo, fromEdbInfo, edbInfo.EdbCode)
		}
		return
	}

	sql := ` DELETE FROM edb_info_calculate_mapping WHERE edb_info_id = ? `
	err = to.Exec(sql, edbInfo.EdbInfoId).Error
	if err != nil {
		return
	}

	sql = ` DELETE FROM ` + tableName + ` WHERE edb_info_id = ? `
	err = to.Exec(sql, edbInfo.EdbInfoId).Error
	if err != nil {
		return
	}

	{
		calculateMappingItem := &EdbInfoCalculateMapping{
			EdbInfoCalculateMappingId: 0,
			EdbInfoId:                 edbInfo.EdbInfoId,
			Source:                    obj.GetSource(),
			SourceName:                obj.GetSourceName(),
			EdbCode:                   edbInfo.EdbCode,
			FromEdbInfoId:             fromEdbInfo.EdbInfoId,
			FromEdbCode:               fromEdbInfo.EdbCode,
			FromEdbName:               fromEdbInfo.EdbName,
			FromSource:                fromEdbInfo.Source,
			FromSourceName:            fromEdbInfo.SourceName,
			FromTag:                   "",
			Sort:                      1,
			CreateTime:                time.Now(),
			ModifyTime:                time.Now(),
			FromSubSource:             fromEdbInfo.SubSource,
		}
		err = to.Create(calculateMappingItem).Error
		if err != nil {
			return
		}
	}

	err, errMsg = obj.refresh(to, edbInfo, fromEdbInfo, edbInfo.EdbCode)

	return
}

func (obj Percentile) Refresh(params RefreshParams) (err error, errMsg string) {
	calculateMapping, err := GetEdbInfoCalculateMappingDetail(params.EdbInfo.EdbInfoId)
	if err != nil {
		errMsg = "GetEdbInfoCalculateLjzzyDetail Err:" + err.Error()
		return
	}
	fromEdbInfo, err := GetEdbInfoById(calculateMapping.FromEdbInfoId)
	if err != nil {
		errMsg = "GetEdbInfoById Err:" + err.Error()
		return
	}

	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	err, errMsg = obj.refresh(to, params.EdbInfo, fromEdbInfo, params.EdbInfo.EdbCode)

	return
}

func (obj Percentile) GetSource() int {
	return utils.DATA_SOURCE_CALCULATE_PERCENTILE
}

func (obj Percentile) GetSourceName() string {
	return utils.DATA_SOURCE_NAME_CALCULATE_PERCENTILE
}

func (obj Percentile) GetEdbType() int {
	return utils.CALCULATE_EDB_TYPE
}

type PercentileConfig struct {
	CalculateValue int    `description:"时间长度期数"`
	CalculateUnit  string `description:"时间长度频度"`
	PercentType    int    `description:"百分位：0-数据区间(兼容历史数据); 1-数据个数；"`
}

func (obj Percentile) refresh(to *gorm.DB, edbInfo, fromEdbInfo *EdbInfo, edbCode string) (err error, errMsg string) {
	edbInfoId := edbInfo.EdbInfoId
	dataTableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
	edbInfoIdStr := strconv.Itoa(edbInfoId)

	var percentileConfig PercentileConfig
	err = json.Unmarshal([]byte(edbInfo.CalculateFormula), &percentileConfig)
	if err != nil {
		return
	}

	fromDataList, err, errMsg := obj.getPercentileData(fromEdbInfo, percentileConfig.CalculateValue, percentileConfig.CalculateUnit, percentileConfig.PercentType)
	if err != nil {
		return
	}

	existDataList, err := GetAllEdbDataListByTo(to, edbInfoId, edbInfo.Source, edbInfo.SubSource)
	if err != nil {
		return
	}
	existDataMap := make(map[string]string)
	removeDataTimeMap := make(map[string]int) //需要移除的日期数据
	for _, v := range existDataList {
		existDataMap[v.DataTime] = v.Value
		removeDataTimeMap[v.DataTime] = 1
	}
	needAddDateMap := make(map[time.Time]int)

	addSql := ` INSERT INTO ` + dataTableName + `(edb_info_id,edb_code,data_time,value,create_time,modify_time,data_timestamp) values `
	var isAdd bool
	for _, tmpData := range fromDataList {
		currDateStr := tmpData.DataTime
		currTime, tmpErr := time.ParseInLocation(utils.FormatDate, currDateStr, time.Local)
		if tmpErr != nil {
			err = tmpErr
			return
		}
		saveValue := decimal.NewFromFloat(tmpData.Value).Round(4).String()

		existVal, ok := existDataMap[currDateStr]
		if ok {
			delete(removeDataTimeMap, currDateStr)

			if existVal != saveValue {
				sql := ` UPDATE %s SET value=?,modify_time=NOW() WHERE edb_info_id=? AND data_time=? `
				sql = fmt.Sprintf(sql, dataTableName)
				err = to.Exec(sql, saveValue, edbInfoId, currDateStr).Error
				if err != nil {
					return
				}
			}

			continue
		}

		timestamp := currTime.UnixNano() / 1e6
		timeStr := fmt.Sprintf("%d", timestamp)
		if _, existOk := needAddDateMap[currTime]; !existOk {
			addSql += GetAddSql(edbInfoIdStr, edbCode, currDateStr, timeStr, saveValue)
			isAdd = true
		}
		needAddDateMap[currTime] = 1
	}

	{
		removeDateList := make([]string, 0)
		for dateTime := range removeDataTimeMap {
			removeDateList = append(removeDateList, dateTime)
		}
		removeNum := len(removeDateList)
		if removeNum > 0 {
			sql := fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? and data_time in (`+utils.GetOrmInReplace(removeNum)+`) `, dataTableName)
			err = to.Exec(sql, edbInfo.EdbInfoId, removeDateList).Error
			if err != nil {
				fmt.Println(reflect.TypeOf(obj).Name(), " add data ;delete Err", err.Error())
				err = fmt.Errorf("删除不存在的指标数据失败,Err:" + err.Error())
				return
			}
		}
	}

	if isAdd {
		addSql = strings.TrimRight(addSql, ",")
		err = to.Exec(addSql).Error
		if err != nil {
			fmt.Println(reflect.TypeOf(obj).Name(), " add data Err", err.Error())
			return
		}
	}

	return
}

func (obj Percentile) calculate(edbInfoId int, date, edbInfoIdStr, edbCode, dataTableName, addSql string, val float64, dataMap, existDataMap map[string]string, to *gorm.DB) (newSql string, isAdd bool, err error) {
	newSql = addSql
	saveValue := utils.SubFloatToString(val, 4)
	if existVal, ok := dataMap[date]; !ok {
		dataTime, _ := time.ParseInLocation(utils.FormatDate, date, time.Local)
		timestamp := dataTime.UnixNano() / 1e6
		timeStr := fmt.Sprintf("%d", timestamp)
		if _, existOk := existDataMap[date]; !existOk {
			newSql += GetAddSql(edbInfoIdStr, edbCode, date, timeStr, saveValue)
			isAdd = true
		}
		existDataMap[date] = date
	} else {
		if existVal != saveValue {
			sql := ` UPDATE %s SET value=?,modify_time=NOW() WHERE edb_info_id=? AND data_time=? `
			sql = fmt.Sprintf(sql, dataTableName)
			err = to.Exec(sql, saveValue, edbInfoId, date).Error
		}
	}
	return
}

func (obj Percentile) getPercentileData(fromEdbInfo *EdbInfo, calculateValue int, calculateUnit string, percentType int) (newDataList []EdbInfoSearchData, err error, errMsg string) {
	dataList := make([]*EdbInfoSearchData, 0)
	switch fromEdbInfo.EdbInfoType {
	case 0:
		dataList, err = GetEdbDataListAll(fromEdbInfo.Source, fromEdbInfo.SubSource, FindEdbDataListAllCond{
			EdbInfoId: fromEdbInfo.EdbInfoId,
		}, 1)
	case 1:
		dataList, err = GetPredictEdbDataListAllByStartDate(fromEdbInfo, 1, "")
	default:
		err = errors.New(fmt.Sprint("获取失败,指标base类型异常", fromEdbInfo.EdbInfoType))
		return
	}

	moveUnitDays, ok := utils.FrequencyDaysMap[calculateUnit]
	if !ok {
		errMsg = `错误的周期`
		err = errors.New(errMsg)
		return
	}
	calculateDay := calculateValue * moveUnitDays

	newDataList = make([]EdbInfoSearchData, 0)

	dataMap := make(map[time.Time]float64, 0)
	for _, tmpData := range dataList {
		currDateTime, _ := time.ParseInLocation(utils.FormatDate, tmpData.DataTime, time.Local)
		dataMap[currDateTime] = tmpData.Value
	}

	if percentType == utils.PercentCalculateTypeRange {
		for i, tmpData := range dataList {
			currDateTime, _ := time.ParseInLocation(utils.FormatDate, tmpData.DataTime, time.Local)
			maxVal := tmpData.Value
			minVal := tmpData.Value
			for k := 0; k < calculateDay; k++ {
				preVal, ok2 := dataMap[currDateTime.AddDate(0, 0, -k)]
				if ok2 {
					if preVal > maxVal {
						maxVal = preVal
					}
					if preVal < minVal {
						minVal = preVal
					}
				}
			}

			if maxVal == minVal {
				continue
			}
			tmpV := (tmpData.Value - minVal) / (maxVal - minVal) * 100
			tmpV, _ = decimal.NewFromFloat(tmpV).Round(4).Float64()
			newDataList = append(newDataList, EdbInfoSearchData{
				EdbDataId: i,
				DataTime:  dataList[i].DataTime,
				Value:     tmpV,
			})
		}
	}
	maxDay := len(dataList) // 往前找数据的边界
	if percentType == utils.PercentCalculateTypeNum {
		for i, d := range dataList {
			s2 := decimal.NewFromFloat(d.Value)
			t2, _ := time.ParseInLocation(utils.FormatDate, d.DataTime, time.Local)

			var bigN, tinyN int
			for k := 0; k < maxDay; k++ {
				if bigN >= calculateDay {
					break
				}
				preVal, preOk := dataMap[t2.AddDate(0, 0, -k)]
				if !preOk {
					continue
				}
				bigN += 1
				if decimal.NewFromFloat(preVal).LessThanOrEqual(s2) {
					tinyN += 1
				}
			}

			if bigN <= 1 {
				continue
			}
			numerator := decimal.NewFromInt(int64(tinyN - 1))
			denominator := decimal.NewFromInt(int64(bigN - 1))
			percentVal, _ := numerator.Div(denominator).Mul(decimal.NewFromFloat(100)).Round(4).Float64()

			newDataList = append(newDataList, EdbInfoSearchData{
				EdbDataId: i,
				DataTime:  dataList[i].DataTime,
				Value:     percentVal,
			})
		}
	}

	return
}
