package models

import (
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"gorm.io/gorm"

	"github.com/shopspring/decimal"
	"strconv"
	"strings"
	"time"
)

func SavePredictCalculateJp(reqEdbInfoId, classifyId int, edbName, frequency, unit, formula string, fromEdbInfo *EdbInfo, edbCode, uniqueCode string, sysUserId int, sysUserRealName, lang string) (edbInfo *EdbInfo, latestDateStr string, latestValue float64, err error, errMsg string) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()
	fmt.Println("reqEdbInfoId:", reqEdbInfoId)

	if reqEdbInfoId <= 0 {
		edbInfo = &EdbInfo{
			EdbInfoType:   1,
			SourceName:    utils.DATA_SOURCE_NAME_PREDICT_CALCULATE_JP,
			Source:        utils.DATA_SOURCE_PREDICT_CALCULATE_JP,
			EdbCode:       edbCode,
			EdbName:       edbName,
			EdbNameSource: edbName,
			Frequency:     frequency,
			Unit:          unit,
			ClassifyId:       classifyId,
			SysUserId:        sysUserId,
			SysUserRealName:  sysUserRealName,
			UniqueCode:       uniqueCode,
			CreateTime:       time.Now(),
			ModifyTime:       time.Now(),
			MinValue:         0,
			MaxValue:         0,
			CalculateFormula: formula,
			EdbType:          2,
			Sort:             GetAddEdbMaxSortByClassifyId(classifyId, utils.PREDICT_EDB_INFO_TYPE),
			MoveType:         0,
			MoveFrequency:    "",
			NoUpdate:         0,
			ServerUrl:        "",
			EdbNameEn:        edbName,
			UnitEn:           unit,
			LatestDate:       "",
			LatestValue:      0,
			ChartImage:       "",
		}
		tmpErr := to.Create(edbInfo).Error
		if tmpErr != nil {
			err = tmpErr
			return
		}

		{
			calculateMappingItem := &EdbInfoCalculateMapping{
				EdbInfoCalculateMappingId: 0,
				EdbInfoId:                 edbInfo.EdbInfoId,
				Source:                    edbInfo.Source,
				SourceName:                edbInfo.SourceName,
				EdbCode:                   edbCode,
				FromEdbInfoId:             fromEdbInfo.EdbInfoId,
				FromEdbCode:               fromEdbInfo.EdbCode,
				FromEdbName:               fromEdbInfo.EdbName,
				FromSource:                fromEdbInfo.Source,
				FromSourceName:            fromEdbInfo.SourceName,
				FromTag:                   "",
				Sort:                      1,
				CreateTime:                time.Now(),
				ModifyTime:                time.Now(),
			}
			err = to.Create(calculateMappingItem).Error
			if err != nil {
				return
			}
		}
	} else {
		edbInfo, err = GetEdbInfoById(reqEdbInfoId)
		if err != nil {
			if err.Error() == utils.ErrNoRow() {
				errMsg = `获取指标信息失败`
			}
			return
		}
		if edbInfo.EdbInfoType != 1 {
			errMsg = `该指标不是预测指标`
			err = errors.New(errMsg)
			return
		}

		latestDateStr = edbInfo.LatestDate
		latestValue = edbInfo.LatestValue
		oldCalculateFormula := edbInfo.CalculateFormula
		switch lang {
		case utils.EnLangVersion:
			edbInfo.EdbNameEn = edbName
			edbInfo.UnitEn = unit
		default:
			edbInfo.EdbName = edbName
			edbInfo.Unit = unit
			edbInfo.EdbNameSource = edbName
		}
		edbInfo.Frequency = frequency
		edbInfo.ClassifyId = classifyId
		edbInfo.CalculateFormula = formula
		edbInfo.ModifyTime = time.Now()
		err = to.Model(edbInfo).Select([]string{"EdbName", "EdbNameSource", "Frequency", "Unit", "ClassifyId", "CalculateFormula", "ModifyTime", "EdbNameEn", "UnitEn"}).Updates(edbInfo).Error
		if err != nil {
			return
		}

		var existCondition string
		var existPars []interface{}
		existCondition += " AND edb_info_id=? AND from_edb_info_id=? "
		existPars = append(existPars, edbInfo.EdbInfoId, fromEdbInfo.EdbInfoId)

		count, tmpErr := GetEdbInfoCalculateCountByCondition(existCondition, existPars)
		if tmpErr != nil {
			err = errors.New("判断指标是否改变失败,Err:" + tmpErr.Error())
			return
		}
		if count > 0 && formula == oldCalculateFormula { // 指标未被替换，无需重新计算
			return
		}

		sql := ` DELETE FROM edb_info_calculate_mapping WHERE edb_info_id = ? `
		err = to.Exec(sql, edbInfo.EdbInfoId).Error
		if err != nil {
			err = errors.New("删除计算指标关联关系失败,Err:" + err.Error())
			return
		}

		tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
		sql = ` DELETE FROM ` + tableName + ` WHERE edb_info_id = ? `
		err = to.Exec(sql, edbInfo.EdbInfoId).Error
		if err != nil {
			return
		}

		{
			calculateMappingItem := &EdbInfoCalculateMapping{
				EdbInfoCalculateMappingId: 0,
				EdbInfoId:                 edbInfo.EdbInfoId,
				Source:                    edbInfo.Source,
				SourceName:                edbInfo.SourceName,
				EdbCode:                   edbCode,
				FromEdbInfoId:             fromEdbInfo.EdbInfoId,
				FromEdbCode:               fromEdbInfo.EdbCode,
				FromEdbName:               fromEdbInfo.EdbName,
				FromSource:                fromEdbInfo.Source,
				FromSourceName:            fromEdbInfo.SourceName,
				FromTag:                   "",
				Sort:                      1,
				CreateTime:                time.Now(),
				ModifyTime:                time.Now(),
			}
			err = to.Create(calculateMappingItem).Error
			if err != nil {
				return
			}
		}

	}

	latestDateStr, latestValue, err = refreshAllPredictCalculateJp(to, edbInfo.EdbInfoId, edbInfo.Source, edbInfo.SubSource, fromEdbInfo, edbCode, edbInfo.Frequency, formula)

	return
}

func RefreshAllPredictCalculateJp(edbInfoId, source, subSource int, fromEdbInfo *EdbInfo, edbCode, edbFrequency, formula string) (latestDateStr string, latestValue float64, err error) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	latestDateStr, latestValue, err = refreshAllPredictCalculateJp(to, edbInfoId, source, subSource, fromEdbInfo, edbCode, edbFrequency, formula)
	return
}

func refreshAllPredictCalculateJp(to *gorm.DB, edbInfoId, source, subSource int, fromEdbInfo *EdbInfo, edbCode, edbFrequency, formula string) (latestDateStr string, latestValue float64, err error) {
	edbInfoIdStr := strconv.Itoa(edbInfoId)
	fmt.Println("EdbInfoId:", fromEdbInfo.EdbInfoId)
	dataList, err := GetPredictEdbDataListAll(fromEdbInfo, 1)
	if err != nil {
		return
	}
	var dateArr []string
	dataMap := make(map[string]*EdbInfoSearchData)
	fromDataMap := make(map[string]float64)
	for _, v := range dataList {
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
		fromDataMap[v.DataTime] = v.Value
	}
	fmt.Println("source:", source)

	existDataList, err := GetAllEdbDataListByTo(to, edbInfoId, source, subSource)
	if err != nil {
		return
	}
	existDataMap := make(map[string]*EdbData)
	existDelDateMap := make(map[string]string)
	for _, v := range existDataList {
		existDataMap[v.DataTime] = v
		existDelDateMap[v.DataTime] = v.DataTime
	}

	latestDateStr = fromEdbInfo.LatestDate

	tableName := GetEdbDataTableName(utils.DATA_SOURCE_PREDICT_CALCULATE_JP, subSource)
	addSql := ` INSERT INTO ` + tableName + ` (edb_info_id,edb_code,data_time,value,create_time,modify_time,data_timestamp) values `
	var isAdd bool

	dataLen := len(dataList)
	if dataLen <= 0 {
		return
	}
	startDataTime, _ := time.ParseInLocation(utils.FormatDate, dataList[0].DataTime, time.Local)
	endDataTime, _ := time.ParseInLocation(utils.FormatDate, dataList[dataLen-1].DataTime, time.Local)

	nextEndDate := utils.GetFrequencyEndDay(startDataTime, edbFrequency) // 下一个节点的日期
	weekDayDataList := make([]float64, 0)
	for currStartDataTime := startDataTime; !currStartDataTime.After(endDataTime); currStartDataTime = currStartDataTime.AddDate(0, 0, 1) {
		if tmpData, ok := dataMap[currStartDataTime.Format(utils.FormatDate)]; ok {
			tmpValue := decimal.NewFromFloat(tmpData.Value)
			tmpValueFloat, _ := tmpValue.Round(4).Float64()
			weekDayDataList = append(weekDayDataList, tmpValueFloat)
		}

		switch edbFrequency {
		case "周度":
			if currStartDataTime.Weekday() != 0 {
				continue
			} else {
				nextEndDate = currStartDataTime.AddDate(0, 0, 7)
			}
		case "旬度":
			nextDay := currStartDataTime.AddDate(0, 0, 1)
			if nextDay.Day() != 1 && nextDay.Day() != 11 && nextDay.Day() != 21 {
				continue
			} else {
				if nextDay.Day() == 1 || nextDay.Day() == 11 {
					nextEndDate = nextDay.AddDate(0, 0, 9)
				} else {
					tmpNextMonth := nextDay.AddDate(0, 1, 0)
					nextEndDate = time.Date(tmpNextMonth.Year(), tmpNextMonth.Month(), 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
				}
			}
		case "月度":
			nextDay := currStartDataTime.AddDate(0, 0, 1)
			if nextDay.Day() != 1 {
				continue
			} else {
				nextEndDate = nextDay.AddDate(0, 1, -1)
			}
		case "季度":
			nextDay := currStartDataTime.AddDate(0, 0, 1)
			if (nextDay.Month() == 1 || nextDay.Month() == 4 || nextDay.Month() == 7 || nextDay.Month() == 10) && nextDay.Day() == 1 {
				nextEndDate = nextDay.AddDate(0, 3, -1)
			} else {
				continue
			}
		case "年度":
			if currStartDataTime.Month() == 12 && currStartDataTime.Day() == 31 {
				nextEndDate = currStartDataTime.AddDate(1, 0, 0)
			} else {
				continue
			}
		default:
			err = errors.New("错误的频度:" + edbFrequency)
			return
		}

		lenWeekDayDataList := len(weekDayDataList)
		if lenWeekDayDataList <= 0 {
			continue
		}

		var currVal float64
		if formula == "期末值" { // 期末值，取区间最后一个日期的数据值
			currVal = weekDayDataList[lenWeekDayDataList-1]
		} else {
			sumValDeci := decimal.NewFromFloat(0)
			for _, v := range weekDayDataList {
				tmpValDeci := decimal.NewFromFloat(v)
				sumValDeci = sumValDeci.Add(tmpValDeci)
			}
			lenDeci := decimal.NewFromInt(int64(lenWeekDayDataList))
			currVal, _ = sumValDeci.Div(lenDeci).Round(4).Float64()
		}

		if fromEdbInfo.LatestDate == currStartDataTime.Format(utils.FormatDate) {
			latestValue = currVal
		}
		currStartDataTimeStr := currStartDataTime.Format(utils.FormatDate)

		if existData, ok := existDataMap[currStartDataTimeStr]; ok {
			existValStr := existData.Value
			existValDeci, tmpErr := decimal.NewFromString(existValStr)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			existVal, _ := existValDeci.Round(4).Float64()
			if existVal != currVal {
				err = ModifyEdbDataById(source, subSource, existData.EdbDataId, fmt.Sprint(currVal))
				if err != nil {
					return
				}
			}

			delete(existDelDateMap, currStartDataTimeStr)
		} else {
			timestamp := currStartDataTime.UnixNano() / 1e6
			timestampStr := fmt.Sprintf("%d", timestamp)
			addSql += GetAddSql(edbInfoIdStr, edbCode, currStartDataTime.Format(utils.FormatDate), timestampStr, fmt.Sprint(currVal))
			isAdd = true

			delete(existDelDateMap, currStartDataTimeStr)
		}

		weekDayDataList = make([]float64, 0)
	}

	lenWeekDayDataList := len(weekDayDataList)
	if lenWeekDayDataList > 0 {
		var currVal float64
		if formula == "期末值" {
			currVal = weekDayDataList[lenWeekDayDataList-1]
		} else {
			sumValDeci := decimal.NewFromFloat(0)
			for _, v := range weekDayDataList {
				tmpValDeci := decimal.NewFromFloat(v)
				sumValDeci = sumValDeci.Add(tmpValDeci)
			}
			lenDeci := decimal.NewFromInt(int64(lenWeekDayDataList))
			currVal, _ = sumValDeci.Div(lenDeci).Round(4).Float64()
		}

		nextEndDateStr := nextEndDate.Format(utils.FormatDate)

		if existData, ok := existDataMap[nextEndDateStr]; ok {
			existValStr := existData.Value
			existValDeci, tmpErr := decimal.NewFromString(existValStr)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			existVal, _ := existValDeci.Round(4).Float64()
			if existVal != currVal {
				err = ModifyEdbDataById(source, subSource, existData.EdbDataId, fmt.Sprint(currVal))
				if err != nil {
					return
				}
			}

			delete(existDelDateMap, nextEndDateStr)
		} else {
			timestamp := nextEndDate.UnixNano() / 1e6
			timestampStr := fmt.Sprintf("%d", timestamp)
			addSql += GetAddSql(edbInfoIdStr, edbCode, nextEndDate.Format(utils.FormatDate), timestampStr, fmt.Sprint(currVal))
			isAdd = true

			delete(existDelDateMap, nextEndDateStr)
		}
	}

	if isAdd {
		addSql = strings.TrimRight(addSql, ",")
		err = to.Exec(addSql).Error
	}

	if len(existDelDateMap) > 0 {
		removeDateList := make([]string, 0) //需要移除的日期
		for k := range existDelDateMap {
			removeDateList = append(removeDateList, k)
		}
		removeDateStr := strings.Join(removeDateList, `","`)
		removeDateStr = `"` + removeDateStr + `"`
		sql := fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? and data_time in (%s) `, tableName, removeDateStr)
		err = to.Exec(sql, edbInfoId).Error
		if err != nil {
			err = fmt.Errorf("删除年化指标数据失败,Err:" + err.Error())
			return
		}
	}

	return
}
