package models

import (
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"github.com/shopspring/decimal"
	"gorm.io/gorm"
	"sort"
	"strings"
	"time"
)

type EdbDataPredictCalculateLjztbpj struct {
	EdbDataId     int       `gorm:"primaryKey;autoIncrement;column:edb_data_id"`
	EdbInfoId     int       `gorm:"column:edb_info_id" description:"指标ID"`
	EdbCode       string    `gorm:"column:edb_code" description:"指标编码"`
	DataTime      string    `gorm:"column:data_time" description:"数据日期"`
	Value         float64   `gorm:"column:value" description:"数据值"`
	Status        int       `gorm:"column:status" description:"状态"`
	CreateTime    time.Time `gorm:"column:status" description:"创建时间"`
	ModifyTime    time.Time `gorm:"column:create_time" description:"修改时间"`
	DataTimestamp int64     `gorm:"column:modify_time" description:"数据日期时间戳"`
}

func RefreshAllPredictCalculateLjztbpj(edbInfo *EdbInfo) (latestDateStr string, latestValue float64, err error) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	var existCondition string
	var existPars []interface{}
	existCondition += " AND edb_info_id=? "
	existPars = append(existPars, edbInfo.EdbInfoId)
	existList, err := GetEdbInfoCalculateListByCondition(existCondition, existPars)
	if err != nil {
		err = errors.New("判断指标是否改变失败,Err:" + err.Error())
		return
	}

	var existItemA, existItemB *EdbInfoCalculateMapping
	for _, existItem := range existList {
		if existItem.FromTag == "A" {
			existItemA = existItem
		} else if existItem.FromTag == "B" {
			existItemB = existItem
		}
	}
	if existItemA == nil {
		err = errors.New("原待拼接指标不存在")
		return
	}
	if existItemB == nil {
		err = errors.New("原同比值指标不存在")
		return
	}
	fromEdbInfo, err := GetEdbInfoById(existItemA.FromEdbInfoId)
	if err != nil {
		err = fmt.Errorf("GetEdbInfoById Err:" + err.Error())
		return
	}

	secondEdbInfo, err := GetEdbInfoById(existItemB.FromEdbInfoId)
	if err != nil {
		err = fmt.Errorf("GetEdbInfoById Err:" + err.Error())
		return
	}
	latestDateStr, latestValue, err = refreshAllPredictCalculateLjztbpj(to, edbInfo, fromEdbInfo, secondEdbInfo, existItemB)
	return
}

func refreshAllPredictCalculateLjztbpj(to *gorm.DB, edbInfo, firstEdbInfo, secondEdbInfo *EdbInfo, existItemB *EdbInfoCalculateMapping) (latestDateStr string, latestValue float64, err error) {
	var dataList []*EdbDataPredictCalculateLjztbpj
	sql := ` SELECT * FROM edb_data_predict_calculate_ljztbpj WHERE edb_info_id=? ORDER BY data_time DESC `
	err = to.Raw(sql, edbInfo.EdbInfoId).Scan(&dataList).Error
	if err != nil {
		return
	}
	latestDateStr = secondEdbInfo.LatestDate

	pjDataMap := make(map[string]float64)     //需要入库的数据
	nowEdbDataMap := make(map[string]float64) //当前指标的数据（已经在库里了，不需要重新）
	pjEdbDataTimeList := make([]string, 0)

	dataMap := make(map[string]*EdbDataPredictCalculateLjztbpj)
	for _, v := range dataList {
		pjEdbDataTimeList = append(pjEdbDataTimeList, v.DataTime)
		dataMap[v.DataTime] = v
		nowEdbDataMap[v.DataTime] = v.Value
	}

	startCalculationDate, _ := time.ParseInLocation(utils.FormatDate, edbInfo.CalculateFormula, time.Local)

	{
		var firstDataList []*EdbInfoSearchData
		firstDataList, err = GetPredictEdbDataListAllByStartDate(firstEdbInfo, 0, "")
		if err != nil {
			return
		}

		for _, v := range firstDataList {
			if v.DataTime > startCalculationDate.Format(utils.FormatDate) {
				continue
			}
			if edbData, ok := dataMap[v.DataTime]; ok {
				if edbData.Value != v.Value {
					edbData.Value = v.Value
					_ = to.Model(edbData).Select([]string{"Value"}).Updates(edbData).Error
				}
			} else {
				pjDataMap[v.DataTime] = v.Value
				pjEdbDataTimeList = append(pjEdbDataTimeList, v.DataTime)
			}

			nowEdbDataMap[v.DataTime] = v.Value
		}
	}

	tbzEdbDataMap := make(map[string]float64)

	tbzEdbDataTimeList := make([]string, 0)

	{
		secondDataList, tmpErr := GetEdbDataListAllByTo(to, existItemB.FromSource, existItemB.FromSubSource, FindEdbDataListAllCond{
			EdbInfoId:         existItemB.FromEdbInfoId,
			StartDataTime:     startCalculationDate.Format(utils.FormatDate),
			StartDataTimeCond: ">",
		}, 0)
		if tmpErr != nil {
			err = tmpErr
			return
		}

		for _, v := range secondDataList {
			tbzEdbDataMap[v.DataTime] = v.Value
			tbzEdbDataTimeList = append(tbzEdbDataTimeList, v.DataTime)
		}
	}

	sort.Strings(tbzEdbDataTimeList)

	removeDateList := make([]string, 0)
	for nowEdbDate := range nowEdbDataMap {
		nowEdbDateTime, _ := time.ParseInLocation(utils.FormatDate, nowEdbDate, time.Local)
		if startCalculationDate.Before(nowEdbDateTime) {
			if _, ok := tbzEdbDataMap[nowEdbDate]; !ok {
				removeDateList = append(removeDateList, nowEdbDate)
			}
		}
	}

	updateEdbDataMap := make(map[string]float64)
	for _, v := range tbzEdbDataTimeList {
		tbzDataTime, _ := time.ParseInLocation(utils.FormatDate, v, time.Local)

		var pjDataTime time.Time
		if tbzDataTime.Month() == 2 {
			pjDataTime = tbzDataTime.AddDate(0, -11, 0)
			pjDataTime = time.Date(pjDataTime.Year(), pjDataTime.Month(), 1, 0, 0, 0, 0, time.Now().Location()).AddDate(0, 0, -1)
		} else {
			pjDataTime = tbzDataTime.AddDate(-1, 0, 0)
		}

		if nowEdbDataValue, isHas := nowEdbDataMap[v]; isHas {
			if lastYearEdbDataValue, ok := nowEdbDataMap[pjDataTime.Format(utils.FormatDate)]; ok {
				tbzDataValue := tbzEdbDataMap[v] //同比值
				currValue := lastYearEdbDataValue * (1 + tbzDataValue/100)
				currValue, _ = decimal.NewFromFloat(currValue).Truncate(4).Float64() //保留4位小数
				if nowEdbDataValue != currValue {
					updateEdbDataMap[v] = currValue
				}
			}
		} else {
			if pjDataValue, ok := nowEdbDataMap[pjDataTime.Format(utils.FormatDate)]; ok {
				tbzDataValue := tbzEdbDataMap[v] //同比值
				currValue := pjDataValue * (1 + tbzDataValue/100)

				currValue, _ = decimal.NewFromFloat(currValue).Truncate(4).Float64()
				nowEdbDataMap[v] = currValue

				pjDataMap[v] = currValue
				pjEdbDataTimeList = append(pjEdbDataTimeList, v)
			}
		}
	}

	{
		addDataList := make([]*EdbDataPredictCalculateLjztbpj, 0)
		sort.Strings(pjEdbDataTimeList)
		for _, dataTime := range pjEdbDataTimeList {
			if dataValue, ok := pjDataMap[dataTime]; ok {
				currentDate, _ := time.ParseInLocation(utils.FormatDate, dataTime, time.Local)
				timestamp := currentDate.UnixNano() / 1e6

				edbDataLjztbpj := &EdbDataPredictCalculateLjztbpj{
					EdbInfoId:     edbInfo.EdbInfoId,
					EdbCode:       edbInfo.EdbCode,
					DataTime:      dataTime,
					Value:         dataValue,
					Status:        1,
					CreateTime:    time.Now(),
					ModifyTime:    time.Now(),
					DataTimestamp: timestamp,
				}
				addDataList = append(addDataList, edbDataLjztbpj)
			}
		}

		if len(addDataList) > 0 {
			tmpErr := to.CreateInBatches(addDataList, 500).Error
			if tmpErr != nil {
				err = tmpErr
				return
			}
		}
	}

	{
		if len(removeDateList) > 0 {
			removeDateStr := strings.Join(removeDateList, `","`)
			removeDateStr = `"` + removeDateStr + `"`
			tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
			sql := fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? and data_time in (%s) `, tableName, removeDateStr)

			err = to.Exec(sql, edbInfo.EdbInfoId).Error
			if err != nil {
				err = errors.New("删除不存在的累计值同比拼接指标数据失败,Err:" + err.Error())
				return
			}
		}
	}

	{
		tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
		for edbDate, edbDataValue := range updateEdbDataMap {
			sql := fmt.Sprintf(` UPDATE  %s set value = ?,modify_time=now() WHERE edb_info_id = ? and data_time = ? `, tableName)
			err = to.Exec(sql, edbDataValue, edbInfo.EdbInfoId, edbDate).Error
			if err != nil {
				err = errors.New("更新现有的累计值同比拼接指标数据失败,Err:" + err.Error())
				return
			}
		}
	}

	var finalLast EdbInfoSearchData
	sql = ` SELECT data_time , value FROM edb_data_predict_calculate_ljztbpj WHERE edb_info_id=? and data_time<=? ORDER BY data_time DESC `
	tmpErr := to.Raw(sql, edbInfo.EdbInfoId, latestDateStr).First(&finalLast).Error
	if tmpErr != nil {
		if tmpErr.Error() != utils.ErrNoRow() {
			err = tmpErr
		}
		return
	} else {
		latestDateStr = finalLast.DataTime
		latestValue = finalLast.Value
	}
	return
}
