package models

import (
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"github.com/shopspring/decimal"
	"gorm.io/gorm"
	"reflect"
	"strconv"
	"strings"
	"time"
)

type PredictTb struct {
}

func (obj PredictTb) Add(params BatchSaveCalculateBatchParams) (edbInfo *EdbInfo, latestDateStr string, latestValue float64, err error, errMsg string) {
	req := params.Req
	fromEdbInfo := params.FromEdbInfo
	edbCode := params.EdbCode

	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	edbInfo = &EdbInfo{
		SourceName:       obj.GetSourceName(),
		Source:           obj.GetSource(),
		EdbCode:          edbCode,
		EdbName:          req.EdbName,
		EdbNameSource:    req.EdbName,
		Frequency:        req.Frequency,
		Unit:             req.Unit,
		StartDate:        "",
		EndDate:          "",
		ClassifyId:       req.ClassifyId,
		SysUserId:        params.SysUserId,
		SysUserRealName:  params.SysUserRealName,
		UniqueCode:       params.UniqueCode,
		CreateTime:       time.Now(),
		ModifyTime:       time.Now(),
		MinValue:         0,
		MaxValue:         0,
		CalculateFormula: req.Formula,
		EdbType:          2,
		Sort:             GetAddEdbMaxSortByClassifyId(req.ClassifyId, utils.PREDICT_EDB_INFO_TYPE),
		MoveType:         0,
		MoveFrequency:    "",
		NoUpdate:         0,
		ServerUrl:        "",
		EdbInfoType:      1,
		EdbNameEn:        req.EdbName,
		UnitEn:           req.Unit,
		LatestDate:       "",
		LatestValue:      0,
		ChartImage:       "",
		Calendar:         "",
	}

	tmpErr := to.Create(edbInfo).Error
	if tmpErr != nil {
		err = tmpErr
		return
	}

	{
		calculateMappingItem := new(EdbInfoCalculateMapping)
		calculateMappingItem.CreateTime = time.Now()
		calculateMappingItem.ModifyTime = time.Now()
		calculateMappingItem.Sort = 1
		calculateMappingItem.EdbCode = edbCode
		calculateMappingItem.EdbInfoId = edbInfo.EdbInfoId
		calculateMappingItem.FromEdbInfoId = fromEdbInfo.EdbInfoId
		calculateMappingItem.FromEdbCode = fromEdbInfo.EdbCode
		calculateMappingItem.FromEdbName = fromEdbInfo.EdbName
		calculateMappingItem.FromSource = fromEdbInfo.Source
		calculateMappingItem.FromSourceName = fromEdbInfo.SourceName
		calculateMappingItem.FromTag = ""
		calculateMappingItem.Source = edbInfo.Source
		calculateMappingItem.SourceName = edbInfo.SourceName
		err = to.Create(calculateMappingItem).Error
		if err != nil {
			return
		}
	}

	latestDateStr, latestValue, err = obj.refresh(to, edbInfo, fromEdbInfo, edbInfo.EdbCode, "")

	return
}

func (obj PredictTb) Edit(params BatchSaveCalculateBatchParams) (latestDateStr string, latestValue float64, err error, errMsg string) {
	req := params.Req
	edbInfo := params.EdbInfo
	fromEdbInfo := params.FromEdbInfo

	latestDateStr = edbInfo.LatestDate
	latestValue = edbInfo.LatestValue

	if fromEdbInfo.Frequency != req.Frequency {
		errMsg = "生成指标频度与原指标频度不同"
		err = errors.New(errMsg)
		return
	}
	if edbInfo.EdbInfoType != 1 {
		errMsg = `该指标不是预测指标`
		err = errors.New(errMsg)
		return
	}

	latestDateStr = edbInfo.LatestDate
	latestValue = edbInfo.LatestValue

	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()
	tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)

	switch params.Lang {
	case utils.EnLangVersion:
		edbInfo.EdbNameEn = req.EdbName
		edbInfo.UnitEn = req.Unit
	default:
		edbInfo.EdbName = req.EdbName
		edbInfo.Unit = req.Unit
		edbInfo.EdbNameSource = req.EdbName
	}
	edbInfo.Frequency = req.Frequency
	edbInfo.ClassifyId = req.ClassifyId
	edbInfo.ModifyTime = time.Now()
	err = to.Model(edbInfo).Select([]string{"EdbName", "EdbNameSource", "Frequency", "Unit", "ClassifyId", "ModifyTime", "EdbNameEn", "UnitEn"}).Updates(edbInfo).Error
	if err != nil {
		return
	}

	var existCondition string
	var existPars []interface{}
	existCondition += " AND edb_info_id=? AND from_edb_info_id=? "
	existPars = append(existPars, edbInfo.EdbInfoId, fromEdbInfo.EdbInfoId)

	var count int
	count, err = GetEdbInfoCalculateCountByCondition(existCondition, existPars)
	if err != nil {
		err = errors.New("判断指标是否改变失败,Err:" + err.Error())
		return
	}
	if count > 0 { // 指标未被替换，无需重新计算
		return
	}
	sql := ` DELETE FROM edb_info_calculate_mapping WHERE edb_info_id = ? `
	err = to.Exec(sql, edbInfo.EdbInfoId).Error
	if err != nil {
		return
	}

	sql = ` DELETE FROM ` + tableName + ` WHERE edb_info_id = ? `
	err = to.Exec(sql, edbInfo.EdbInfoId).Error
	if err != nil {
		return
	}

	{
		calculateMappingItem := &EdbInfoCalculateMapping{
			EdbInfoCalculateMappingId: 0,
			EdbInfoId:                 edbInfo.EdbInfoId,
			Source:                    obj.GetSource(),
			SourceName:                obj.GetSourceName(),
			EdbCode:                   edbInfo.EdbCode,
			FromEdbInfoId:             fromEdbInfo.EdbInfoId,
			FromEdbCode:               fromEdbInfo.EdbCode,
			FromEdbName:               fromEdbInfo.EdbName,
			FromSource:                fromEdbInfo.Source,
			FromSourceName:            fromEdbInfo.SourceName,
			FromTag:                   "",
			Sort:                      1,
			CreateTime:                time.Now(),
			ModifyTime:                time.Now(),
		}
		err = to.Create(calculateMappingItem).Error
		if err != nil {
			return
		}
	}

	latestDateStr, latestValue, err = obj.refresh(to, edbInfo, fromEdbInfo, edbInfo.EdbCode, "")

	return
}

func (obj PredictTb) Refresh(params RefreshParams) (latestDateStr string, latestValue float64, err error, errMsg string) {
	calculateMapping, err := GetEdbInfoCalculateMappingDetail(params.EdbInfo.EdbInfoId)
	if err != nil {
		return
	}
	fromEdbInfo, err := GetEdbInfoById(calculateMapping.FromEdbInfoId)
	if err != nil {
		errMsg = "GetEdbInfoById Err:" + err.Error()
		return
	}

	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()

	latestDateStr, latestValue, err = obj.refresh(to, params.EdbInfo, fromEdbInfo, params.EdbInfo.EdbCode, params.StartDate)

	return
}

func (obj PredictTb) GetSource() int {
	return utils.DATA_SOURCE_PREDICT_CALCULATE_TBZ
}

func (obj PredictTb) GetSourceName() string {
	return utils.DATA_SOURCE_NAME_PREDICT_CALCULATE_TBZ
}

func (obj PredictTb) refresh(to *gorm.DB, edbInfo, fromEdbInfo *EdbInfo, edbCode, startDate string) (latestDateStr string, latestValue float64, err error) {
	edbInfoId := edbInfo.EdbInfoId
	dataTableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)

	edbInfoIdStr := strconv.Itoa(edbInfoId)

	dataList, err := GetPredictEdbDataListAllByStartDate(fromEdbInfo, 0, "")
	if err != nil {
		return
	}

	calculateDataList, err := EdbInfoSearchDataToEdbInfoData(dataList)
	if err != nil {
		return
	}
	baseCalculate := BaseCalculate{
		DataList:  calculateDataList,
		Frequency: fromEdbInfo.Frequency,
		Source:    3,
	}
	dateDataMap, err, _ := baseCalculate.Tbz()
	if err != nil {
		return
	}

	latestDateStr = fromEdbInfo.LatestDate

	existDataList, err := GetAllEdbDataListByTo(to, edbInfoId, edbInfo.Source, edbInfo.SubSource)
	if err != nil {
		return
	}
	existDataMap := make(map[string]string)
	removeDataTimeMap := make(map[string]int) //需要移除的日期数据
	for _, v := range existDataList {
		existDataMap[v.DataTime] = v.Value
		removeDataTimeMap[v.DataTime] = 1
	}

	addSql := ` INSERT INTO edb_data_predict_calculate_tbz (edb_info_id,edb_code,data_time,value,create_time,modify_time,data_timestamp) values `
	var isAdd bool
	existAddDataMap := make(map[string]string)

	for currentDate, valFloat := range dateDataMap {
		currentDateStr := currentDate.Format(utils.FormatDate)
		timestamp := currentDate.UnixNano() / 1e6
		timestampStr := fmt.Sprintf("%d", timestamp)
		val := decimal.NewFromFloat(valFloat).String()

		delete(removeDataTimeMap, currentDateStr)

		if existVal, ok := existDataMap[currentDateStr]; !ok {
			if _, existOk := existAddDataMap[currentDateStr]; !existOk {
				addSql += GetAddSql(edbInfoIdStr, edbCode, currentDateStr, timestampStr, val)
				isAdd = true
			}
			existAddDataMap[currentDateStr] = currentDateStr
		} else {
			existValDecimal, tmpErr := decimal.NewFromString(existVal)
			if tmpErr != nil {
				err = tmpErr
				return
			}
			existStr := existValDecimal.String()
			if existStr != val {
				sql := ` UPDATE %s SET value=?,modify_time=NOW() WHERE edb_info_id=? AND data_time=? `
				sql = fmt.Sprintf(sql, dataTableName)
				err = to.Exec(sql, val, edbInfoId, currentDateStr).Error
				if err != nil {
					return
				}
			}
		}

		if fromEdbInfo.LatestDate == currentDateStr {
			latestValueDecimal, tmpErr := decimal.NewFromString(val)
			if tmpErr != nil {
				return
			}
			latestValue, _ = latestValueDecimal.Truncate(4).Float64() //保留4位小数
		}
	}

	{
		removeDateList := make([]string, 0)
		for dateTime := range removeDataTimeMap {
			removeDateList = append(removeDateList, dateTime)
		}
		removeNum := len(removeDateList)
		if removeNum > 0 {
			sql := fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? and data_time in (`+utils.GetOrmInReplace(removeNum)+`) `, dataTableName)
			err = to.Exec(sql, edbInfo.EdbInfoId, removeDateList).Error
			if err != nil {
				fmt.Println(reflect.TypeOf(obj).Name(), " add data ;delete Err", err.Error())
				err = fmt.Errorf("删除不存在的指标数据失败,Err:" + err.Error())
				return
			}
		}
	}

	if isAdd {
		addSql = strings.TrimRight(addSql, ",")
		err = to.Exec(addSql).Error
		if err != nil {
			return
		}
	}

	{
		finalLast, tmpErr := GetFinalLastByTo(to, edbInfoId, edbInfo.Source, edbInfo.SubSource, fromEdbInfo.LatestDate)
		if tmpErr != nil && tmpErr.Error() != utils.ErrNoRow() {
			return
		}
		if tmpErr == nil {
			latestDateStr = finalLast.DataTime
			latestValue = finalLast.Value
		}
	}

	return
}
