package models

import (
	"errors"
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"gorm.io/gorm"
	"strings"
	"time"
)

type EdbDataPredictCalculateZjpj struct {
	EdbDataId     int `gorm:"primaryKey;autoIncrement;column:edb_data_id"`
	EdbInfoId     int
	EdbCode       string
	DataTime      string
	Value         float64
	Status        int
	CreateTime    time.Time
	ModifyTime    time.Time
	DataTimestamp int64
}

func RefreshAllPredictCalculateZjpj(edbInfo *EdbInfo) (latestDateStr string, latestValue float64, err error) {
	to := global.DEFAULT_DmSQL.Begin()
	defer func() {
		if err != nil {
			to.Rollback()
		} else {
			to.Commit()
		}
	}()
	var existCondition string
	var existPars []interface{}
	existCondition += " AND edb_info_id=? "
	existPars = append(existPars, edbInfo.EdbInfoId)
	existList, err := GetEdbInfoCalculateListByCondition(existCondition, existPars)
	if err != nil {
		err = fmt.Errorf("判断指标是否改变失败,Err:" + err.Error())
		return
	}

	var existItemA, existItemB *EdbInfoCalculateMapping
	for _, existItem := range existList {
		if existItem.FromTag == "A" {
			existItemA = existItem
		} else if existItem.FromTag == "B" {
			existItemB = existItem
		}
	}
	if existItemA == nil {
		err = errors.New("原拼接日期之前的指标不存在")
		return
	}
	if existItemB == nil {
		err = errors.New("原拼接日期之后的指标不存在")
		return
	}

	fromEdbInfo, err := GetEdbInfoById(existItemA.FromEdbInfoId)
	if err != nil {
		err = fmt.Errorf("GetEdbInfoById Err:" + err.Error())
		return
	}

	secondEdbInfo, err := GetEdbInfoById(existItemB.FromEdbInfoId)
	if err != nil {
		err = fmt.Errorf("GetEdbInfoById Err:" + err.Error())
		return
	}
	latestDateStr, latestValue, err = refreshAllPredictCalculateZjpj(to, edbInfo, fromEdbInfo, secondEdbInfo)

	return
}

func refreshAllPredictCalculateZjpj(to *gorm.DB, edbInfo, firstEdbInfo, secondEdbInfo *EdbInfo) (latestDateStr string, latestValue float64, err error) {
	var dataList []*EdbDataPredictCalculateZjpj
	sql := ` SELECT * FROM edb_data_predict_calculate_zjpj WHERE edb_info_id=? ORDER BY data_time DESC `
	err = to.Raw(sql, edbInfo.EdbInfoId).Scan(&dataList).Error
	if err != nil {
		return
	}
	if edbInfo.CalculateFormula <= secondEdbInfo.LatestDate {
		latestDateStr = secondEdbInfo.LatestDate
	} else {
		if edbInfo.CalculateFormula >= firstEdbInfo.LatestDate {
			latestDateStr = firstEdbInfo.LatestDate
		} else {
			latestDateStr = edbInfo.CalculateFormula
		}
	}
	var dateArr []string
	dataMap := make(map[string]*EdbDataPredictCalculateZjpj)
	removeDataTimeMap := make(map[string]int) //需要移除的日期数据
	for _, v := range dataList {
		dateArr = append(dateArr, v.DataTime)
		dataMap[v.DataTime] = v
		removeDataTimeMap[v.DataTime] = 1
	}

	addDataList := make([]*EdbDataPredictCalculateZjpj, 0)
	{
		var firstDataList []*EdbInfoSearchData
		firstDataList, err = GetPredictEdbDataListAllByStartDate(firstEdbInfo, 0, "")
		if err != nil {
			return
		}

		for _, v := range firstDataList {
			if v.DataTime >= edbInfo.CalculateFormula {
				continue
			}
			if _, ok := removeDataTimeMap[v.DataTime]; ok {
				delete(removeDataTimeMap, v.DataTime)
			}
			if latestDateStr == v.DataTime {
				latestValue = v.Value
			}
			if edbData, ok := dataMap[v.DataTime]; ok {
				if edbData.Value != v.Value {
					edbData.Value = v.Value
					_ = to.Model(edbData).Select([]string{"Value"}).Updates(edbData).Error
				}
			} else {
				currentDate, _ := time.ParseInLocation(utils.FormatDate, v.DataTime, time.Local)
				timestamp := currentDate.UnixNano() / 1e6

				edbDataZjpj := &EdbDataPredictCalculateZjpj{
					EdbInfoId:     edbInfo.EdbInfoId,
					EdbCode:       edbInfo.EdbCode,
					DataTime:      v.DataTime,
					Value:         v.Value,
					Status:        1,
					CreateTime:    time.Now(),
					ModifyTime:    time.Now(),
					DataTimestamp: timestamp,
				}
				addDataList = append(addDataList, edbDataZjpj)
			}
		}
	}

	{
		var secondDataList []*EdbInfoSearchData
		secondDataList, err = GetPredictEdbDataListAllByStartDate(secondEdbInfo, 0, edbInfo.CalculateFormula)
		if err != nil {
			return
		}
		for _, v := range secondDataList {
			if _, ok := removeDataTimeMap[v.DataTime]; ok {
				delete(removeDataTimeMap, v.DataTime)
			}
			if latestDateStr == v.DataTime {
				latestValue = v.Value
			}
			if edbData, ok := dataMap[v.DataTime]; ok {
				if edbData.Value != v.Value {
					edbData.Value = v.Value
					edbData.ModifyTime = time.Now()
					tmpErr := to.Model(edbData).Select([]string{"Value", "ModifyTime"}).Updates(edbData).Error
					if tmpErr != nil {
						fmt.Println("tmpErr:", tmpErr)
						err = tmpErr
						return
					}
				}
			} else {
				currentDate, _ := time.ParseInLocation(utils.FormatDate, v.DataTime, time.Local)
				timestamp := currentDate.UnixNano() / 1e6

				edbDataZjpj := &EdbDataPredictCalculateZjpj{
					EdbInfoId:     edbInfo.EdbInfoId,
					EdbCode:       edbInfo.EdbCode,
					DataTime:      v.DataTime,
					Value:         v.Value,
					Status:        1,
					CreateTime:    time.Now(),
					ModifyTime:    time.Now(),
					DataTimestamp: timestamp,
				}
				addDataList = append(addDataList, edbDataZjpj)
			}
		}
	}

	{
		removeDateList := make([]string, 0)
		for dateTime := range removeDataTimeMap {
			removeDateList = append(removeDateList, dateTime)
		}
		if len(removeDateList) > 0 {
			removeDateStr := strings.Join(removeDateList, `","`)
			removeDateStr = `"` + removeDateStr + `"`
			tableName := GetEdbDataTableName(edbInfo.Source, edbInfo.SubSource)
			sql := fmt.Sprintf(` DELETE FROM %s WHERE edb_info_id = ? and data_time in (%s) `, tableName, removeDateStr)

			err = to.Exec(sql, edbInfo.EdbInfoId).Error
			if err != nil {
				err = fmt.Errorf("删除不存在的直接拼接指标数据失败,Err:" + err.Error())
				return
			}
		}
	}

	if len(addDataList) > 0 {
		tmpErr := to.CreateInBatches(addDataList, 500).Error
		if tmpErr != nil {
			err = tmpErr
			return
		}
	}
	return
}
