package models

import (
	"eta_gn/eta_index_lib/global"
	"eta_gn/eta_index_lib/utils"
	"gorm.io/gorm"
	"time"
)

type PredictEdbRuleData struct {
	PredictEdbRuleDataId int `gorm:"primaryKey;autoIncrement;column:predict_edb_rule_data_id"`
	EdbInfoId            int
	ConfigId             int
	DataTime             string
	Value                string
	CreateTime           time.Time
	ModifyTime           time.Time
	DataTimestamp        int64
}

type PredictEdbRuleDataItem struct {
	PredictEdbRuleDataId int `gorm:"primaryKey;autoIncrement;column:predict_edb_rule_data_id"`
	EdbInfoId            int
	ConfigId             int
	DataTime             string
	Value                float64
	CreateTime           time.Time
	ModifyTime           time.Time
	DataTimestamp        int64
}

func GetPredictEdbRuleDataItemList(edbInfoId, configId int, startDate, endDate string) (list []*PredictEdbRuleDataItem, err error) {
	var pars []interface{}
	sql := ` SELECT * FROM predict_edb_rule_data WHERE edb_info_id = ? AND config_id = ? `
	pars = append(pars, edbInfoId, configId)
	if startDate != "" {
		sql += ` AND data_time>=? `
		pars = append(pars, startDate)
	}
	if endDate != "" {
		sql += ` AND data_time<=? `
		pars = append(pars, endDate)
	}
	sql += ` ORDER BY data_time ASC `
	err = global.DEFAULT_DmSQL.Raw(sql, pars...).Scan(&list).Error

	if err != nil {
		return
	}

	for _, v := range list {
		v.ConvertDate()
	}

	return
}

func GetPredictEdbRuleDataByTo(to *gorm.DB, configId int) (list []*PredictEdbRuleData, err error) {
	sql := `SELECT * FROM predict_edb_rule_data WHERE config_id = ?`
	err = to.Raw(sql, configId).Find(&list).Error
	return
}

func (m *PredictEdbRuleData) AfterFind(db *gorm.DB) (err error) {
	if m.DataTime == "" {
		return
	}
	dataDateTmp, err := time.ParseInLocation(utils.FormatDateWallWithLoc, m.DataTime, time.Local)
	if err != nil {
		return
	}
	m.DataTime = dataDateTmp.Format(utils.FormatDate)
	return
}

func (m *PredictEdbRuleData) ConvertDate() {
	m.DataTime = utils.GormDateStrToDateStr(m.DataTime)

	return
}

func (m *PredictEdbRuleDataItem) ConvertDate() {
	m.DataTime = utils.GormDateStrToDateStr(m.DataTime)

	return
}
