package edb_refresh

import (
	"eta_gn/eta_index_lib/models/edb_refresh"
	"eta_gn/eta_index_lib/utils"
	"fmt"
	"time"
)

func GetPreviousHalfHourDefaultConfigList(source, subSource int) (list []*edb_refresh.EdbRefreshDefaultConfig, err error) {
	defer func() {
		if err != nil {
			fmt.Println(err)
		}
	}()
	refreshFrequencyList := []string{"每自然日", "每交易日", "每周", "每旬", "每月", "每季", "每半年", "每年"}
	now := time.Now()

	currTimeStr := getPreviousHalfHour(now)
	fmt.Println(currTimeStr)

	list = make([]*edb_refresh.EdbRefreshDefaultConfig, 0)
	for _, refreshFrequency := range refreshFrequencyList {
		condition, pars, isHandler := getRefreshFrequencyCondition(now, refreshFrequency)
		if !isHandler {
			continue
		}

		condition += ` AND refresh_frequency = ? AND refresh_time = ? AND source = ? AND sub_source = ? `
		pars = append(pars, refreshFrequency, currTimeStr, source, subSource)

		tmpList, tmpErr := edb_refresh.GetListByCondition(condition, pars)
		if tmpErr != nil {
			err = tmpErr
			return
		}
		list = append(list, tmpList...)
	}

	return
}

func mergeMaps(dst map[string][]*edb_refresh.EdbInfoListAndRefreshConfig, src map[string][]*edb_refresh.EdbInfoListAndRefreshConfig) map[string][]*edb_refresh.EdbInfoListAndRefreshConfig {
	if dst == nil {
		return src
	}
	if src == nil {
		return dst
	}
	for k, v := range src {
		if dstk, ok := dst[k]; ok {
			dstk = append(dstk, v...)
			dst[k] = dstk
		} else {
			dst[k] = v
		}
	}
	return dst
}

func getRefreshFrequencyCondition(now time.Time, refreshFrequency string) (condition string, pars []interface{}, isHandler bool) {
	isHandler = true

	var dayNum int
	var isLastDay bool

	switch refreshFrequency {
	case "每自然日":
		return
	case "每交易日":
		if now.Weekday() == time.Saturday || now.Weekday() == time.Sunday {
			isHandler = false
		}
		return
	case "每周":
		currWeekDay := now.Weekday()
		if currWeekDay == time.Sunday {
			currWeekDay = 7
			isLastDay = true
		}
		dayNum = int(currWeekDay)
	case "每旬":
		currDay := now.Day()
		if currDay <= 10 {
			dayNum = currDay
			if currDay == 10 {
				isLastDay = true
			}
		} else if currDay <= 20 {
			dayNum = currDay - 10
			if currDay == 20 {
				isLastDay = true
			}
		} else {
			dayNum = currDay - 20

			monthLastDay := time.Date(now.Year(), now.Month()+1, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
			if currDay == monthLastDay.Day() {
				isLastDay = true
			}
		}
	case "每月":
		currDay := now.Day()
		dayNum = currDay

		monthLastDay := time.Date(now.Year(), now.Month()+1, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)

		if currDay == monthLastDay.Day() {
			isLastDay = true
		}
	case "每季":
		var startDay, endDay time.Time
		currMonth := now.Month()
		currDay := now.Day()
		if currMonth <= 3 {
			startDay = time.Date(now.Year(), 1, 1, 0, 0, 0, 0, time.Local)
			endDay = time.Date(now.Year(), 4, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
		} else if currMonth <= 6 {
			startDay = time.Date(now.Year(), 4, 1, 0, 0, 0, 0, time.Local)
			endDay = time.Date(now.Year(), 7, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
		} else if currMonth <= 9 {
			startDay = time.Date(now.Year(), 7, 1, 0, 0, 0, 0, time.Local)
			endDay = time.Date(now.Year(), 10, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
		} else {
			startDay = time.Date(now.Year(), 10, 1, 0, 0, 0, 0, time.Local)
			endDay = time.Date(now.Year()+1, 1, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
		}

		dayNum = utils.GetTimeSubDay(startDay, now) + 1

		if currMonth == endDay.Month() && currDay == endDay.Day() {
			isLastDay = true
		}
	case "每半年":
		var startDay, endDay time.Time
		currMonth := now.Month()
		currDay := now.Day()
		if currMonth <= 6 {
			startDay = time.Date(now.Year(), 1, 1, 0, 0, 0, 0, time.Local)
			endDay = time.Date(now.Year(), 7, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
		} else {
			startDay = time.Date(now.Year(), 7, 1, 0, 0, 0, 0, time.Local)
			endDay = time.Date(now.Year()+1, 1, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)
		}

		dayNum = utils.GetTimeSubDay(startDay, now) + 1

		if currMonth == endDay.Month() && currDay == endDay.Day() {
			isLastDay = true
		}
	case "每年":
		currMonth := now.Month()
		currDay := now.Day()

		startDay := time.Date(now.Year(), 1, 1, 0, 0, 0, 0, time.Local)
		endDay := time.Date(now.Year()+1, 1, 1, 0, 0, 0, 0, time.Local).AddDate(0, 0, -1)

		dayNum = utils.GetTimeSubDay(startDay, now) + 1

		if currMonth == endDay.Month() && currDay == endDay.Day() {
			isLastDay = true
		}

	}

	if isLastDay {
		condition += ` AND ( refresh_frequency_day = ? OR refresh_frequency_day >= ? )`
		pars = append(pars, 0, dayNum)
	} else {
		condition += ` AND refresh_frequency_day = ?  `
		pars = append(pars, dayNum)
	}

	return
}

func getPreviousHalfHour(now time.Time) string {
	minute := now.Minute()

	if minute >= 30 {
		return fmt.Sprintf("%02d:%02d", now.Hour(), 30)
	}
	return fmt.Sprintf("%02d:%02d", now.Hour(), 0)
}
